# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Optional

from ..damlast.visitor import ExprVisitor, IdentityTypeVisitor
from ..damlast.daml_lf_1 import Expr, Type, PrimType, FieldWithType, PrimCon, PrimLit, BuiltinFunction, ValName
from ..model.types_store import PackageStore
from ..util.typing import safe_cast


class TypeComputer(ExprVisitor[Type], IdentityTypeVisitor):
    """
    Evaluate the :class:`Type` of an :class:`Expr`.
    """
    def __init__(self, store: 'PackageStore'):
        self.store = safe_cast(PackageStore, store)

    def resolve_type(self, var: str) -> 'Optional[Type]':
        return None

    def visit_expr_var(self, var: str) -> 'Type':
        pass

    def visit_expr_val(self, val: 'ValName') -> 'Type':
        vref = self.store.resolve_value_reference(val)
        if vref is not None:
            return self.visit_expr(vref)
        else:
            raise RuntimeError(f'unknown val name: {val}')

    def visit_expr_builtin(self, builtin: 'BuiltinFunction') -> 'Type':
        pass

    def visit_expr_prim_con(self, prim_con: 'PrimCon') -> 'Type':
        if PrimCon.CON_TRUE == prim_con or PrimCon.CON_FALSE == prim_con:
            return Type(prim=Type.Prim(PrimType.BOOL, args=()))
        else:
            return Type(prim=Type.Prim(PrimType.UNIT, args=()))

    def visit_expr_prim_lit(self, prim_lit: 'PrimLit') -> 'Type':
        if prim_lit.int64 is not None:
            return Type(prim=Type.Prim(PrimType.INT64, args=()))
        elif prim_lit.date is not None:
            return Type(prim=Type.Prim(PrimType.DATE, args=()))
        elif prim_lit.timestamp is not None:
            return Type(prim=Type.Prim(PrimType.TIMESTAMP, args=()))
        elif prim_lit.decimal is not None:
            return Type(prim=Type.Prim(PrimType.DECIMAL, args=()))
        elif prim_lit.party is not None:
            return Type(prim=Type.Prim(PrimType.PARTY, args=()))
        elif prim_lit.text is not None:
            return Type(prim=Type.Prim(PrimType.TEXT, args=()))
        else:
            raise RuntimeError('unknown prim lit type')

    def visit_expr_rec_con(self, rec_con: 'Expr.RecCon') -> 'Type':
        return self.visit_type(Type(con=rec_con.tycon))

    def visit_expr_rec_proj(self, rec_proj: 'Expr.RecProj') -> 'Type':
        rec_type = self.visit_type_con(rec_proj.tycon)
        old_type = self.store.resolve_type_reference(rec_type.con.tycon)
        for n, v in old_type.as_args_list():
            if n == rec_proj.field:
                return get_new_type(v)

    def visit_expr_variant_con(self, variant_con: 'Expr.VariantCon') -> 'Type':
        return self.visit_type(Type(con=variant_con.tycon))

    def visit_expr_tuple_con(self, tuple_con: 'Expr.TupleCon') -> 'Type':
        return Type(tuple=Type.Tuple([FieldWithType(fwe.field, self.visit_expr(fwe.expr))
                                      for fwe in tuple_con.fields]))

    def visit_expr_tuple_proj(self, tuple_proj: 'Expr.TupleProj') -> 'Type':
        pass

    def visit_expr_app(self, app: 'Expr.App') -> 'Type':
        fun_type = self.visit_expr(app.fun)
        if fun_type.prim.prim == PrimType.ARROW:
            t = fun_type
            arg_count = len(app.args)
            while arg_count > 0:
                arg_count -= 1
                t = t.prim.args[1]
            return t
        else:
            raise ValueError(f'unexpected type of app body: {fun_type}')

    def visit_expr_ty_app(self, ty_app: 'Expr.TyApp') -> 'Type':
        core_type = self.visit_expr(ty_app.expr)
        if ty_app.types:
            def unexpected_type_error(*_, **__):
                raise RuntimeError('Unexpected tyapp')

            return core_type.Sum_match(
                lambda var: Type(var=Type.Var(var=var.var, args=var.args + ty_app.types)),
                lambda con: Type(con=Type.Con(tycon=con.tycon, args=con.args + ty_app.types)),
                lambda prim: Type(prim=Type.Prim(prim=prim.prim, args=prim.args + ty_app.types)),
                lambda forall: unexpected_type_error,
                lambda tuple: unexpected_type_error)
        return core_type

    def visit_expr_abs(self, abs_: 'Expr.Abs') -> 'Type':
        t = self.visit_expr(abs_.body)
        for param in reversed(abs_.param):
            t = arrow_type(param.type, t)
        return t

    def visit_expr_ty_abs(self, ty_abs: 'Expr.TyAbs') -> 'Type':
        return '...'
        #scope = self.with_type_var() ty_abs.param[0].var
        #return scope.visit_expr(ty_abs.body)

    def visit_expr_case(self, case: 'Case') -> 'Type':
        # TODO: Probably not right
        return self.visit_expr(case.alts[0].body)

    def visit_expr_let(self, let: 'Block') -> 'Type':
        return self.visit_expr(let.body)

    def visit_expr_nil(self, nil: 'Expr.Nil') -> 'Type':
        return Type(prim=Type.Prim(PrimType.LIST, [self.visit_type(nil.type)]))

    def visit_expr_cons(self, cons: 'Expr.Cons') -> 'Type':
        return Type(prim=Type.Prim(PrimType.LIST, [self.visit_type(cons.type)]))

    def visit_expr_update(self, update: 'Update') -> 'Type':
        raise Exception('Update not yet supported')

    def visit_expr_scenario(self, scenario: 'Scenario') -> 'Type':
        raise Exception('Scenario not yet supported')

    def visit_expr_rec_upd(self, rec_upd: 'Expr.RecUpd') -> 'Type':
        raise Exception('Update not yet supported')

    def visit_expr_tuple_upd(self, tuple_upd: 'Expr.TupleUpd') -> 'Type':
        raise Exception('Update not yet supported')

    def visit_expr_none(self, none: 'Expr.None_') -> 'Type':
        return Type(prim=Type.Prim(PrimType.OPTIONAL, [self.visit_type(none.type)]))

    def visit_expr_some(self, some: 'Expr.Some') -> 'Type':
        return Type(prim=Type.Prim(PrimType.OPTIONAL, [self.visit_type(some.type)]))

    def visit_type_var(self, var: 'Type.Var') -> 'Type':
        resolved_type = self.resolve_type(var.var)
        if resolved_type is not None:
            return self.visit_type(resolved_type)
        else:
            return Type(var=var)


def arrow_type(input: 'Type', output: 'Type') -> 'Type':
    return Type(prim=Type.Prim(PrimType.ARROW, (input, output)))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Write-Side types
----------------

The :mod:`dazl.model.writing` module contains the Python classes used on the write-side of the
Ledger API.

.. autoclass:: Command
   :members:

.. autoclass:: CreateCommand
   :members:

.. autoclass:: ExerciseCommand
   :members:
"""
import uuid
import warnings
from datetime import datetime, timedelta
from typing import Any, Collection, Dict, Generic, List, Optional, Sequence, TypeVar, Union

from dataclasses import dataclass, fields

from .. import LOG
from .core import ContractId, Party
from .types import Type, UnresolvedTypeReference, TemplateChoice, RecordType, UnsupportedType, \
    VariantType, ContractIdType, ListType, scalar_type_dispatch_table, \
    TypeEvaluationContext, type_evaluate_dispatch, OptionalType, TemplateMeta, ChoiceMeta, MapType
from .types_store import PackageStore
from ..util.typing import safe_cast

TCommand = TypeVar('TCommand')
TValue = TypeVar('TValue')


CommandsOrCommandSequence = Union[None, 'Command', List[Optional['Command']]]
EventHandlerResponse = Union[CommandsOrCommandSequence, 'CommandBuilder', 'CommandPayload']


class Command:
    """
    Base class for write-side commands.
    """


@dataclass(init=False, frozen=True)
class CreateCommand(Command):
    """
    A command that creates a contract without any predecessors.

    .. attribute:: CreateCommand.template

        Refers to the type of a template. This can be passed in as a ``str`` to the constructor,
        where it assumed to represent the ID or name of a template.

    .. attribute:: CreateCommand.arguments

        The arguments to the create (as a ``dict``).
    """
    __slots__ = ('template', 'arguments')

    template: Type
    arguments: Dict[str, Any]

    def __init__(self, template: Union[str, Type], arguments=None):
        object.__setattr__(self, 'template', template if isinstance(template, Type)
                           else UnresolvedTypeReference(template))
        object.__setattr__(self, 'arguments', arguments or dict())

    def replace(self, template: Union[None, str, Type] = None, arguments=None):
        """
        Create a new :class:`CreateCommand` with the same identifier as this command, but with new
        values for its parameters.

        :param template:
            The new value of the `template` field, or `None` to reuse the existing value.
        :param arguments:
            The new value of the `arguments` field, or `None` to reuse the existing value.
        """
        if template is not None:
            template = template if isinstance(template, Type) \
                else UnresolvedTypeReference(template)
        return CreateCommand(
            template if template is not None else self.template,
            arguments if arguments is not None else self.arguments)

    def __repr__(self):
        return f'<create {self.template} {self.arguments}>'


@dataclass(init=False, frozen=True)
class ExerciseCommand(Command):
    """
    A command that exercises a choice on a pre-existing contract.

    .. attribute:: ExerciseCommand.contract

        The :class:`ContractId` on which a choice is being exercised.

    .. attribute:: ExerciseCommand.choice

        Refers to a choice (either a :class:`ChoiceRef` or a :class:`ChoiceMetadata`).
        This can be passed in as a ``str`` to the constructor, where it assumed to represent the
        name of a choice.

    .. attribute:: ExerciseCommand.arguments

        The arguments to the exercise choice (as a ``dict``).

    Note that when an ``ExerciseCommand`` is created, an additional ``template_id`` parameter can
    be supplied to the constructor to aid in disambiguation of the specific choice being invoked.
    In some situations involving composite commands, a ``template_id`` must eventually be supplied
    before a choice can be exercised. If this ``template_id`` is specified, the ``contract`` and
    ``choice`` are both tagged with this ID.

    Instance methods:

    .. automethod:: replace
    """
    __slots__ = ('contract', 'choice', 'arguments')

    def __init__(self, contract: Union[str, ContractId], choice: str, arguments=None, template_id=None):
        if isinstance(contract, str):
            warnings.warn('Untyped ContractId support will be removed with the removal of '
                          'the deprecated REST API.', DeprecationWarning, stacklevel=2)
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                contract = ContractId(contract, template_id=template_id)
        elif template_id is not None:
            warnings.warn(
                'Specifying template_id in an ExerciseChoice is no longer necessary. ' +
                'Please avoid specifying it, as this parameter will be removed in the future.',
                DeprecationWarning, stacklevel=2)
        elif not isinstance(contract, ContractId):
            raise ValueError('ContractId expected here')

        object.__setattr__(self, 'contract', contract)
        object.__setattr__(self, 'choice', choice)
        object.__setattr__(self, 'arguments', dict(arguments) if arguments is not None else dict())

    def replace(self, contract=None, choice=None, arguments=None, template_id=None):
        """
        Create a new :class:`ExerciseCommand` with the same identifier as this command, but with new
        values for its parameters.

        :param contract:
            The new value of the `contract` field, or `None` to reuse the existing value.
            The same type coercion rules used in the constructor apply here.
        :param choice:
            The new value of the `choice` field, or `None` to reuse the existing value.
            The same type coercion rules used in the constructor apply here.
        :param arguments:
            The new value of the `choice` field, or `None` to reuse the existing value.
        :param template_id:
            The expected template type.
        """
        return ExerciseCommand(
            contract if contract is not None else self.contract,
            choice if choice is not None else self.choice,
            arguments if arguments is not None else self.arguments,
            template_id)

    def __repr__(self):
        return '<exercise \"{}\" with {} on {}>'.format(
            self.choice,
            self.arguments,
            self.contract.contract_id if hasattr(self.contract, 'contract_id') else self.contract)


class CommandBuilder:
    """
    Builder class for generating commands to be sent to the ledger.
    """

    @classmethod
    def coerce(cls, obj, atomic_default=False) -> 'CommandBuilder':
        """
        Create a :class:`CommandBuilder` from the objects that an event handler is allowed to
        return.

        :param obj:
        :param atomic_default:
        :return:
        """
        if isinstance(obj, CommandBuilder):
            return obj

        builder = CommandBuilder(atomic_default=atomic_default)
        if obj is not None:
            builder.append(obj)
        return builder

    def __init__(self, atomic_default=False):
        self._atomic_default = atomic_default
        self._commands = [[]]  # type: List[List[Command]]
        self._defaults = CommandDefaults()

    def defaults(self,
                 party: Optional[Party] = None,
                 ledger_id: Optional[str] = None,
                 workflow_id: Optional[str] = None,
                 application_id: Optional[str] = None,
                 command_id: Optional[str] = None) -> None:
        if party is not None:
            self._defaults.default_party = party
        if ledger_id is not None:
            self._defaults.ledger_id = ledger_id
        if workflow_id is not None:
            self._defaults.default_workflow_id = workflow_id
        if application_id is not None:
            self._defaults.default_application_id = application_id
        if command_id is not None:
            self._defaults.default_command_id = command_id

    def create(self, template, arguments=None) -> 'CommandBuilder':
        return self.append(create(template, arguments=arguments))

    def exercise(self, contract, choice, arguments=None) -> 'CommandBuilder':
        return self.append(exercise(contract, choice, arguments=arguments))

    def append(self, *commands: CommandsOrCommandSequence) -> 'CommandBuilder':
        """
        Append one or more commands, or list of commands to the :class:`CommandBuilder` in flight.
        This method respects the value of ``atomic_default`` that this object was constructed with.
        In order to force commands to be submitted either atomically, use :meth:`append_atomically`.
        To allow these commands to be submitted in parallel use :meth:`append_nonatomically`.

        :param commands: One or more commands, or list of commands to be submitted to the ledger.
        :return: This object.
        """
        if self._atomic_default:
            # a command builder that defaults to being atomic will put all commands in a single
            # transaction; build on the very first transaction
            self._commands[0].extend(flatten_command_sequence(commands))
            return self
        else:
            return self.append_nonatomically(*commands)

    def append_atomically(self, *commands: Union[Command, Sequence[Command]]) -> 'CommandBuilder':
        self._commands.extend([flatten_command_sequence(commands)])
        return self

    def append_nonatomically(self, *commands: Union[Command, Sequence[Command]]) -> \
            'CommandBuilder':
        self._commands.extend([[cmd] for cmd in flatten_command_sequence(commands)])
        return self

    def build(self, defaults: 'Optional[CommandDefaults]' = None, now: Optional[datetime] = None) \
            -> 'Collection[CommandPayload]':
        """
        Return a collection of commands.
        """
        if defaults is None:
            raise ValueError('defaults must currently be specified')

        return [CommandPayload(
            party=defaults.default_party or self._defaults.default_party,
            ledger_id=defaults.default_ledger_id or self._defaults.default_ledger_id,
            workflow_id=defaults.default_workflow_id or self._defaults.default_workflow_id,
            application_id=defaults.default_application_id or self._defaults.default_application_id,
            command_id=defaults.default_command_id or self._defaults.default_command_id or uuid.uuid4().hex,
            ledger_effective_time=now,
            maximum_record_time=now + (defaults.default_ttl or self._defaults.default_ttl),
            commands=commands
        ) for i, commands in enumerate(self._commands) if commands]

    def __format__(self, format_spec):
        if format_spec == 'c':
            return str(self._commands)
        else:
            return repr(self)

    def __repr__(self):
        return f'CommandBuilder({self._commands})'


def flatten_command_sequence(commands: Sequence[CommandsOrCommandSequence]) -> List[Command]:
    """
    Convert a list of mixed commands, ``None``, and list of commands into an ordered sequence of
    non-``None`` commands.
    """
    ret = []  # type: List[Command]
    errors = []

    for i, obj in enumerate(commands):
        if obj is not None:
            if isinstance(obj, Command):
                ret.append(obj)
            else:
                try:
                    cmd_iter = iter(obj)
                except TypeError:
                    errors.append(((i,), obj))
                    continue
                for j, cmd in enumerate(cmd_iter):
                    if isinstance(cmd, Command):
                        ret.append(cmd)
                    else:
                        errors.append(((i, j), cmd))
    if errors:
        raise ValueError(f'Failed to interpret some elements as Commands in the list: '
                         f'$[{index}] = {command}' for index, command in errors)
    return ret


@dataclass
class CommandDefaults:
    """
    Values to use for a :class:`Command` when no value is specified with the creation of the
    command.
    """

    default_party: Optional[Party] = None
    default_ledger_id: Optional[str] = None
    default_workflow_id: Optional[str] = None
    default_application_id: Optional[str] = None
    default_command_id: Optional[str] = None
    default_ttl: Optional[timedelta] = None


@dataclass(frozen=True)
class CommandPayload:
    """
    A request to mutate active state of the ledger.

    .. attribute:: CommandPayload.party
        The party submitting the request.
    .. attribute:: CommandPayload.application_id:
        An optional application ID to accompany the request.
    .. attribute:: CommandPayload.command_id:
        A hash that represents the BIM commitment.
    .. attribute:: CommandPayload.ledger_effective_time:
        The effective time of this command. Should usually be set to ``datetime.now()``, but
        may have a different value when the server is operating in static time mode.
    .. attribute:: CommandPayload.maximum_record_time:
        The maximum time before the client should consider this command expired.
    .. attribute:: CommandPayload.commands
        A sequence of commands to submit to the ledger. These commands are submitted atomically
        (in other words, they all succeed or they all fail).
    """
    party: Party
    ledger_id: str
    workflow_id: str
    application_id: str
    command_id: str
    ledger_effective_time: datetime
    maximum_record_time: datetime
    commands: Sequence[Command]

    def __post_init__(self):
        missing_fields = [field.name for field in fields(self) if getattr(self, field.name) is None]
        if missing_fields:
            raise ValueError(f'Some fields are set to None when they are required: {missing_fields}')


def create(template, arguments=None):
    from .types_dynamic import NamedRecord, ProxyMeta

    template_type = type(template)
    if isinstance(template_type, TemplateMeta):
        # static codegen, instantiated type
        if arguments is not None:
            raise ValueError('arguments cannot be specified with an instantiated template')
        arguments = template._asdict()
        template = str(template_type)

    elif isinstance(template, NamedRecord):
        # dynamic "codegen", instantiated type
        if arguments is not None:
            raise ValueError('arguments cannot be specified with an instantiated template')
        template, arguments = template.name, template.arguments

    elif template_type == TemplateMeta:
        # static codegen, non-instantiated
        template = str(template)

    elif isinstance(template_type, ProxyMeta):
        # dynamic codegen, non-instantiated
        template = str(template_type)

    elif not isinstance(template, str):
        raise ValueError('template must be a string name, a template type, or an instantiated template')

    return CreateCommand(template, arguments)


def exercise(contract, choice, arguments=None):
    from .types_dynamic import NamedRecord, ProxyMeta

    choice_type = type(choice)
    if isinstance(choice_type, ChoiceMeta):
        # static codegen, instantiated type
        if arguments is not None:
            raise ValueError('arguments cannot be specified with an instantiated template')
        arguments = choice._asdict()
        choice = str(choice_type)

    elif isinstance(choice, NamedRecord):
        # dynamic "codegen", instantiated type
        if arguments is not None:
            raise ValueError('arguments cannot be specified with an instantiated template')
        choice, arguments = choice.name, choice.arguments
        choice_start_idx = choice.rfind('.')
        if choice_start_idx >= 0:
            choice = choice[choice_start_idx + 1:]

    elif choice_type == ChoiceMeta:
        # static codegen, non-instantiated
        choice = str(choice)

    elif isinstance(choice_type, ProxyMeta):
        # dynamic codegen, non-instantiated
        choice = str(choice_type)
        choice_start_idx = choice.rfind('.')
        if choice_start_idx >= 0:
            choice = choice[choice_start_idx + 1:]

    elif not isinstance(choice, str):
        raise ValueError('choice must be a string name, a template type, or an instantiated template')

    return ExerciseCommand(contract, choice, arguments)


####################################################################################################
# argument iteration support
####################################################################################################


def arg_iter(value):
    """
    Produce an iterator that walks over an argument tree and all of its values, recursing into
    lists and record/variant fields.
    """
    if isinstance(value, str):
        # str is a common case, and it's also iterable (which we don't want to exploit here)
        yield value
    elif isinstance(value, dict):
        for sub_value in value.values():
            yield sub_value
    elif hasattr(value, '__iter__'):
        for sub_value in value:
            yield sub_value
    else:
        yield value


class Serializer(Generic[TCommand, TValue]):
    """
    Serializer interface for objects on the write-side of the API.
    """

    def serialize_value(self, type_token: Type, obj: Any) -> TValue:
        raise NotImplementedError('serialize_value requires an implementation')

    def serialize_command(self, command: Command) -> TCommand:
        raise NotImplementedError('serialize_command requires an implementation')


class AbstractSerializer(Serializer[TCommand, TValue]):
    """
    Implementation of :class:`Serializer` that helps enforce that all possible cases of type
    serialization have been implemented.
    """
    def __init__(self, store: PackageStore):
        self.store = safe_cast(PackageStore, store)

    def serialize_value(self, tt: Type, obj: Any) -> TValue:
        context = TypeEvaluationContext.from_store(self.store)
        try:
            return self._serialize_dispatch(context, tt, obj)
        except:
            from ..util.fmt_py import python_example_object

            LOG.warning("Expected something like:")
            for line in str.splitlines(python_example_object(self.store, tt)):
                LOG.warning('    %s', line)

            LOG.warning("But got this instead:")
            LOG.warning('    %r', obj)
            raise

    def serialize_commands(self, commands: Sequence[Command]) -> Sequence[TCommand]:
        return [self.serialize_command(cmd) for cmd in commands]

    def serialize_command(self, command: Command) -> TCommand:
        if isinstance(command, CreateCommand):
            template = command.template
            candidates = self.store.resolve_template_type(template)
            if len(candidates) == 0:
                msg = f'Could not resolve {template} to any valid types'
                LOG.error(msg)
                raise ValueError(msg)
            elif len(candidates) > 1:
                msg = f'Could not uniquely resolve {template} to a single valid type'
                LOG.error(msg)
                raise ValueError(msg)

            tt, = candidates.values()
            if not isinstance(tt, RecordType):
                msg = f'CreateCommand requires a type that is a record (got {tt} instead)'
                LOG.error(msg)
                raise ValueError(msg)

            value = self.serialize_value(tt, command.arguments)
            return self.serialize_create_command(tt, value)
        elif isinstance(command, ExerciseCommand):
            template_type_ref = command.contract.template_id
            choice_name = command.choice
            choice_opts = self.store.resolve_choice(template_type_ref, choice_name)
            if len(choice_opts) == 0:
                msg = f'Could not resolve {template_type_ref} {choice_name} to any valid choices'
                LOG.error(msg)
                raise ValueError(msg)
            if len(choice_opts) > 1:
                msg = f'Could not uniquely resolve {template_type_ref} {choice_name} ' \
                    f'to a single valid choice'
                LOG.error(msg)
                raise ValueError(msg)
            tt, choice = next(iter(choice_opts.items()))

            args = self.serialize_value(choice.type, command.arguments)
            return self.serialize_exercise_command(command.contract, choice, args)

    def serialize_create_command(
            self, template_type: RecordType, template_args: TValue) \
            -> TCommand:
        raise NotImplementedError('serialize_create_command requires an implementation')

    def serialize_exercise_command(
            self, contract_id: ContractId, choice_info: TemplateChoice, choice_args: TValue) \
            -> TCommand:
        raise NotImplementedError('serialize_exercise_command requires an implementation')

    def serialize_unit(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_unit requires an implementation')

    def serialize_bool(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_text(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_int(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_decimal(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_party(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_date(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_datetime(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_timedelta(self, context: TypeEvaluationContext, obj: Any) -> TValue:
        raise NotImplementedError('serialize_bool requires an implementation')

    def serialize_contract_id(self, context: TypeEvaluationContext, tt: ContractIdType, obj: Any) \
            -> TValue:
        raise NotImplementedError('serialize_contract_id requires an implementation')

    def serialize_optional(self, context: TypeEvaluationContext, tt: OptionalType, obj: Any) \
            -> TValue:
        raise NotImplementedError('serialize_optional requires an implementation')

    def serialize_list(self, context: TypeEvaluationContext, tt: ListType, obj: Any) -> TValue:
        raise NotImplementedError('serialize_list requires an implementation')

    def serialize_map(self, context: TypeEvaluationContext, tt: MapType, obj: Any) -> TValue:
        raise NotImplementedError('serialize_map requires an implementation')

    def serialize_record(self, context: TypeEvaluationContext, tt: RecordType, obj: Any) -> TValue:
        raise NotImplementedError('serialize_record requires an implementation')

    def serialize_variant(self, context: TypeEvaluationContext, tt: VariantType, obj: Any) \
            -> TValue:
        raise NotImplementedError('serialize_variant requires an implementation')

    def serialize_unsupported(self, context: TypeEvaluationContext, tt: UnsupportedType, obj: Any) \
            -> TValue:
        raise NotImplementedError('serialize_unsupported requires an implementation')

    def _serialize_dispatch(self, context: TypeEvaluationContext, tt: Type, obj: Any) -> TValue:
        eval_fn = type_evaluate_dispatch(
            lambda c, st: scalar_type_dispatch_table(
                lambda: self.serialize_unit(c, obj),
                lambda: self.serialize_bool(c, obj),
                lambda: self.serialize_text(c, obj),
                lambda: self.serialize_int(c, obj),
                lambda: self.serialize_decimal(c, obj),
                lambda: self.serialize_party(c, obj),
                lambda: self.serialize_date(c, obj),
                lambda: self.serialize_datetime(c, obj),
                lambda: self.serialize_timedelta(c, obj))(st),
            lambda c, ct: self.serialize_contract_id(c, ct, obj),
            lambda c, ot: self.serialize_optional(c, ot, obj),
            lambda c, lt: self.serialize_list(c, lt, obj),
            lambda c, mt: self.serialize_map(c, mt, obj),
            lambda c, rt: self.serialize_record(c, rt, obj),
            lambda c, vt: self.serialize_variant(c, vt, obj),
            lambda c, ut: self.serialize_unsupported(c, ut, obj))
        return eval_fn(context, tt)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
The :mod:`dazl.util.prim_types` module contains functions for converting arbitrary objects and
on-wire formats to/from canonical Python types.
"""
from decimal import Decimal

from datetime import datetime, timezone, date, timedelta
from functools import partial
from typing import overload, Any, Mapping, Tuple, Union


TimeDeltaConvertible = Union[int, float, Decimal, str, timedelta]

DATETIME_ISO8601_Z_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


@overload
def to_boolean(obj: 'Union[bool, str]') -> bool: ...


@overload
def to_boolean(obj: None) -> None: ...


def to_boolean(obj) -> bool:
    """
    Convert any of the common wire representations of a ``bool`` to a ``bool``.
    """
    if obj is None:
        # noinspection PyTypeChecker
        return None

    if isinstance(obj, bool):
        return obj
    elif isinstance(obj, str):
        lobj = obj.lower().strip()
        if lobj == 'true':
            return True
        elif lobj == 'false':
            return False
        else:
            raise ValueError(f"Could not parse a string as a boolean: {obj!r}")
    raise ValueError(f'Could not parse as a boolean: {obj!r}')


@overload
def to_timedelta(obj: TimeDeltaConvertible) -> timedelta: ...


@overload
def to_timedelta(obj: None) -> None: ...


def to_timedelta(obj) -> timedelta:
    """
    Convert an object to a timedelta. If the object is a timedelta, it is returned; if it is a
    numeric type, the value is interpreted as seconds and that is returned.

    :param obj: The object to convert.
    :return: A timedelta.

    >>> to_timedelta(timedelta(minutes=1, seconds=2))
    timedelta(minutes=1, seconds=2)
    >>> to_timedelta(30)
    timedelta(seconds=30)
    >>> to_timedelta(30.0)
    timedelta(seconds=30)
    >>> to_timedelta(Decimal(30))
    timedelta(seconds=30)
    """
    if obj is None:
        raise ValueError('obj cannot be None')
    if isinstance(obj, timedelta):
        return obj
    elif isinstance(obj, (int, float)):
        return timedelta(seconds=obj)
    elif isinstance(obj, Decimal):
        return timedelta(seconds=float(obj))
    elif isinstance(obj, str):
        return timedelta(seconds=float(obj))
    else:
        raise TypeError('could not convert {obj!r} to timedelta')


class PrimitiveTypeConverter:
    to_boolean = staticmethod(to_boolean)

    def to_int(self, obj: Any) -> int:
        if obj is None:
            # noinspection PyTypeChecker
            return None

        if isinstance(obj, int):
            return obj
        elif isinstance(obj, str):
            return int(obj)
        elif isinstance(obj, Decimal):
            return int(obj)
        raise ValueError(f'Could not parse as an int: {obj!r}')

    def to_decimal(self, obj: Any) -> Decimal:
        """
        Convert any of the common wire representations of a ``Decimal`` to a ``Decimal``.
        """
        if obj is None:
            # noinspection PyTypeChecker
            return None

        if isinstance(obj, Decimal):
            return obj
        elif isinstance(obj, (int, float)):
            return Decimal(str(obj))
        elif isinstance(obj, str):
            return Decimal(obj)
        raise ValueError(f'Could not parse as a Decimal: {obj!r}')

    def to_str(self, obj: Any) -> str:
        if obj is None:
            # noinspection PyTypeChecker
            return None

        if isinstance(obj, str):
            return obj
        else:
            return str(obj)

    def to_date(self, obj: Any) -> date:
        """
        Convert any of the common wire representations of a ``date`` to a ``date``.
        """
        if obj is None:
            # noinspection PyTypeChecker
            return None

        if isinstance(obj, date):
            return obj
        elif isinstance(obj, datetime):
            return obj.date()
        elif isinstance(obj, (int, float)):
            return date(1970, 1, 1) + timedelta(days=obj)
        elif isinstance(obj, str):
            for fmt in _KNOWN_DATE_FORMATS:
                try:
                    return fmt(obj).date()
                except ValueError:
                    pass

        raise ValueError(f'Could not parse as a date: {obj!r}')

    def to_datetime(self, obj: Any) -> datetime:
        """
        Convert any of the common wire representations of a ``datetime`` to a ``datetime``.
        """
        try:
            # noinspection PyPackageRequirements
            from google.protobuf.timestamp_pb2 import Timestamp
        except ImportError:
            # noinspection PyPep8Naming
            Timestamp = None

        if obj is None:
            # noinspection PyTypeChecker
            return None

        if isinstance(obj, datetime):
            # trivially return datetime objects untouched
            return obj
        elif Timestamp is not None and isinstance(obj, Timestamp):
            # straight from a Google Timestamp
            dt = obj.ToDatetime()
            return dt.replace(tzinfo=timezone.utc)
        elif isinstance(obj, (int, float)):
            # from the gRPC Ledger API; interpret as microseconds from the GMT epoch
            return datetime.utcfromtimestamp(obj / 1e6).replace(tzinfo=timezone.utc)
        elif isinstance(obj, str):
            # from the REST Ledger API or an end user; give preference to the unambiguous wire
            # format ISO8601, but otherwise try to parse in a variety of formats
            for fmt in _KNOWN_TIME_FORMATS:
                try:
                    return fmt(obj)
                except ValueError:
                    pass

        raise ValueError(f'Could not parse as a datetime: {obj!r}')

    to_timedelta = staticmethod(to_timedelta)


def decode_variant_dict(obj: Any) -> Tuple[str, Any]:
    if not isinstance(obj, dict):
        raise ValueError(f'cannot coerce {obj!r} to a variant')
    if len(obj) != 1:
        raise ValueError(f'variants must be encoded as single-key dicts (got {obj!r} instead)')
    key = list(obj)[0]
    value = obj.get(key)
    return key, value


def standard_format(fmt, value):
    return datetime.strptime(value, fmt).replace(tzinfo=timezone.utc)


def nano_format(value):
    if value.endswith('Z'):
        return standard_format(DATETIME_ISO8601_Z_FORMAT, value[0:-4] + 'Z')
    else:
        raise ValueError('could not parse')


def unflatten_dotted_keys(m: Mapping[str, Any]):
    if m is None:
        return None

    # pull out any specialized dotted-field mappings
    reformatted = dict()
    for key, value in m.items():
        k1, d, k2 = key.partition('.')
        if d:
            v = reformatted.get(k1)
            if v is None:
                v = dict()
                reformatted[k1] = v
            v[k2] = value
        else:
            reformatted[key] = value

    return reformatted


_GLOBAL_TYPE_CONVERTER = PrimitiveTypeConverter()
to_boolean = _GLOBAL_TYPE_CONVERTER.to_boolean
to_int = _GLOBAL_TYPE_CONVERTER.to_int
to_decimal = _GLOBAL_TYPE_CONVERTER.to_decimal
to_str = _GLOBAL_TYPE_CONVERTER.to_str
to_date = _GLOBAL_TYPE_CONVERTER.to_date
to_datetime = _GLOBAL_TYPE_CONVERTER.to_datetime


_KNOWN_DATE_FORMATS = [
    partial(standard_format, '%Y-%m-%d'),
    partial(standard_format, '%Y.%m.%d'),
]


_KNOWN_TIME_FORMATS = [
    partial(standard_format, DATETIME_ISO8601_Z_FORMAT),
    nano_format,
    partial(standard_format, '%Y-%m-%dT%H:%M:%S.%f'),
    partial(standard_format, '%Y-%m-%d %H:%M:%S.%f'),
    partial(standard_format, '%Y-%m-%dT%H:%M:%SZ'),
    partial(standard_format, '%Y-%m-%dT%H:%M:%S'),
    partial(standard_format, '%Y-%m-%d %H:%M:%S'),
]


def frozendict(*args, **kwargs):
    """
    Create a :class:`FrozenDict`, a special subclass of `dict` that is immutable and hashable.
    """
    class HackyDict(dict):
        def __hash__(self):
            return 0

        def __delitem__(self, *args, **kwargs):
            raise RuntimeError('frozendicts are immutable')

        def __setitem__(self, key, value):
            raise RuntimeError('frozendicts are immutable')

        def pop(self, *args, **kwargs):
            raise RuntimeError('frozendicts are immutable')

        def popitem(self, *args, **kwargs):
            raise RuntimeError('frozendicts are immutable')

        def setdefault(self, *args, **kwargs):
            raise RuntimeError('frozendicts are immutable')

        def update(self, *args, **kwargs):
            raise RuntimeError('frozendicts are immutable')

        def clear(self, *args, **kwargs):
            raise RuntimeError('frozendicts are immutable')

    return HackyDict(*args, **kwargs)


def to_hashable(obj):
    from collections import Mapping, Collection
    if isinstance(obj, Mapping):
        return frozendict(obj)
    elif isinstance(obj, str):
        return obj
    elif isinstance(obj, Collection):
        return tuple(obj)
    else:
        return obj



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from asyncio import ensure_future, gather, get_event_loop, Future, InvalidStateError
from functools import wraps

from dataclasses import replace, dataclass, field, fields
from datetime import datetime, timedelta
from io import StringIO
from typing import Any, Awaitable, Callable, Collection, List, Optional, Sequence, Set, Tuple, \
    TypeVar
import reprlib
import uuid

from .. import LOG
from .config import NetworkConfig, PartyConfig
from ._writer_verify import ValidateSerializer
from ..client.state import ActiveContractSet
from ..metrics import MetricEvents
from ..model.core import ContractMatch, ContractsState, ContractId, ContractData, \
    CommandTimeoutError, Party, ContractContextualData, ContractContextualDataCollection
from ..model.ledger import LedgerMetadata
from ..model.network import connection_settings
from ..model.reading import BaseEvent, TransactionStartEvent, TransactionEndEvent, OffsetEvent, \
    TransactionFilter, EventKey, ContractCreateEvent, ContractArchiveEvent, \
    sortable_offset_height, InitEvent, ActiveContractSetEvent
from ..model.writing import CommandBuilder, CommandDefaults, CommandPayload, \
    EventHandlerResponse, Command
from ..protocols import LedgerNetwork, LedgerClient
from ..util.asyncio_util import ServiceQueue, await_then, completed, propagate, \
    safe_create_future, named_gather, failed, Invoker
from ..util.events import CallbackManager
from ..util.prim_natural import n_things
from ..util.typing import safe_cast

T = TypeVar('T')


class _PartyClientImpl:
    def __init__(self, metrics: 'MetricEvents', invoker: 'Invoker', party: 'Party'):
        self.metrics = metrics
        self.invoker = invoker
        self.party = party

        self._config_values = dict()
        self._config = None  # type: Optional[NetworkConfig]
        self._pool = None  # type: Optional[LedgerNetwork]
        self._pool_fut = None  # type: Optional[Awaitable[LedgerNetwork]]
        self._client_fut = None  # type: Optional[Awaitable[LedgerClient]]
        self._ready_fut = safe_create_future()

        self._acs = ActiveContractSet()
        self._reader = _PartyClientReaderState()
        self._writer = _PartyClientWriterState()

        self._callbacks = CallbackManager()

    def connect_in(self, pool: LedgerNetwork, base_config: dict) -> Future:
        LOG.debug('Starting party %r with base config: %r and party config: %r', self.party, base_config, self._config_values)
        base_party_config = _find_party_config(base_config, self.party) or dict()

        config_names = {f.name for f in fields(PartyConfig)}
        all_config = {**base_config, **base_party_config, **self._config_values}

        self._config = config = PartyConfig(**{k: v for k, v in all_config.items() if k in config_names})
        settings, url_prefix = connection_settings(
            config.url,
            self.party,
            verify_ssl=config.verify_ssl,
            ca_file=config.ca_file,
            cert_file=config.cert_file,
            cert_key_file=config.cert_key_file)

        self._pool = pool
        self._client_fut = ensure_future(pool.connect(
            self.party, settings, url_prefix, config.admin_url))
        return self._client_fut

    def set_config(self, **kwargs):
        self._config_values.update(kwargs)

    def initialize(self, current_time: datetime, metadata: LedgerMetadata) -> Awaitable[None]:
        """
        Initialize the state of the ledger.

        :param current_time:
            The ledger time to publish as part of the :class:`InitEvent`.
        :param metadata:
            Information about the connected ledger.
        :return:
        """
        self._acs.metadata_future.set_result(metadata.store)
        evt = InitEvent(self, self.party, current_time, metadata.ledger_id, metadata.store)
        return self.emit_event(evt)

    def ready(self) -> Awaitable[None]:
        return self._ready_fut

    def get_time(self) -> Awaitable[datetime]:
        return await_then(self._pool.ledger(), lambda metadata: metadata.time_model.get_time())

    def set_time(self, new_datetime) -> Awaitable[None]:
        return await_then(self._pool, lambda pool: pool.set_time(new_datetime))

    # region Event Handler Management

    # noinspection PyShadowingBuiltins
    def add_event_handler(self, key, handler, filter, context):
        from functools import wraps

        @wraps(handler)
        def rewritten_event(event):
            new_event = replace(event, client=context)
            return handler(new_event)

        h = wrap_as_command_submission(self.write_commands, rewritten_event, filter)
        self._callbacks.add_listener(key, h)

    def emit_event(self, data: BaseEvent) -> Awaitable[Any]:
        """
        Emit an event.

        :param data:
            The event to raise.
        :return:
            An Awaitable that is resolved when the commands that resulted from event callbacks
            have been completed either successfully or unsuccessfully.
        """
        futures = [ensure_future(fut)
                   for key in EventKey.from_event(data)
                   for fut in self._callbacks.for_listeners(key)(data)]
        LOG.debug('The event %s yielded futures %s', data, futures)
        if len(futures) == 0:
            return completed(None)
        elif len(futures) == 1:
            return futures[0]
        else:
            return named_gather(repr(data), *futures, return_exceptions=True)

    # endregion

    # region Active/Historical Contract Set management

    def find_by_id(self, cid: ContractId) -> ContractContextualData:
        return self._acs.get(cid)

    def find(self,
             template: Any,
             match: ContractMatch = None,
             include_archived: bool = False) \
            -> ContractContextualDataCollection:
        return self._acs.read_full(template, match, include_archived=include_archived)

    def find_active(self, template: Any, match: ContractMatch = None) -> ContractsState:
        return self._acs.read_active(template, match)

    def find_historical(self, template: Any, match: ContractMatch = None) \
            -> ContractContextualDataCollection:
        return self._acs.read_full(template, match, include_archived=True)

    def find_nonempty(self, template: Any, match: ContractMatch, min_count: int = 1,
                      timeout: float = 30):
        return self._acs.read_async(template, match, min_count=min_count)

    # endregion

    # region Read Path

    async def read_acs(self, until_offset: 'Optional[str]', raise_events: bool) \
            -> 'Tuple[str, Future]':
        """
        Initial bootstrap of events from the read side. Only one instance of this coroutine
        should be active at a time per client. An initial block of events is read using the
        Active Contract Set service, and the transaction stream is then subscribed to in order to
        ensure that this client is caught up to the specified offset.

        :param until_offset:
            The destination ledger offset to read up until to. If not included, then the client
            attempts to read as many transactions as is currently available.
        :param raise_events:
            ``True`` to raise transaction- and contract-related events to event handlers;
            ``False`` to suppress this behavior and only update local state within the client
            (offset information and active contract set).
        :return:
            The ledger offset that the reader ended at and a Future that is resolved when all event
            handlers' follow-ups have either successfully or unsuccessfully completed.
        """
        LOG.info('Calling read_acs(%r, %r)', until_offset, raise_events)
        client = await self._client_fut
        acs = await client.active_contracts()
        for acs_evt in acs:
            await self._process_transaction_stream_event(acs_evt, False)

        return await self.read_transactions(until_offset, raise_events)

    async def read_transactions(self, until_offset: 'Optional[str]', raise_events: bool) \
            -> 'Tuple[str, Future]':
        """
        Main processing method of events from the read side. Only one instance of this coroutine
        should be active at a time per client.

        :param until_offset:
            The destination ledger offset to read up until to. If not included, then the client
            attempts to read as many transactions as is currently available.
        :param raise_events:
            ``True`` to raise transaction- and contract-related events to event handlers;
            ``False`` to suppress this behavior and only update local state within the client
            (offset information and active contract set).
        :return:
            The ledger offset that the reader ended at and a Future that is resolved when all event
            handlers' follow-ups have either successfully or unsuccessfully completed.
        """
        LOG.verbose('  read_transactions(%s, %s) with party %s | (groups %s), reader offset %s',
                    until_offset, raise_events, self.party, self._config.party_groups,
                    self._reader.offset)
        client = await self._client_fut
        metadata = await self._pool.ledger()

        event_count = 0
        futs = []

        while (until_offset is None or
               self._reader.offset is None or
               self._reader.offset != until_offset):
            LOG.verbose('  read_transactions(%s, %s) with party %s | current offset: %s; '
                        'destination offset: %s',
                        until_offset, raise_events, self.party, self._reader.offset, until_offset)

            # prepare a call to /events
            transaction_filter = TransactionFilter(
                ledger_id=metadata.ledger_id,
                current_offset=self._reader.offset,
                destination_offset=until_offset,
                templates=None,
                max_blocks=None,
                party_groups=self._config.party_groups)

            transaction_events = await client.events(transaction_filter)
            for event in transaction_events:
                futs.append(self._process_transaction_stream_event(event, raise_events))
            event_count += len(transaction_events)

            # if a destination offset is not specified, then only read one pass of transaction
            # events before exiting
            if until_offset is None:
                break

        self.metrics.party_offset(self.party, self._reader.offset)
        LOG.verbose('  read_transactions(%s, %s) with party %s | reader offset %s',
                    until_offset, raise_events, self.party, self._reader.offset)
        futs = [fut for fut in futs if not fut.done()]
        if len(futs) == 0:
            return self._reader.offset, completed(None)
        elif len(futs) == 1:
            return self._reader.offset, futs[0]
        else:
            return self._reader.offset, gather(*futs, return_exceptions=True)

    async def read_end(self) -> str:
        client = await self._client_fut
        return await client.events_end()

    def _process_transaction_stream_event(self, event: Any, raise_events: bool) -> Future:
        """

        :param event:
        :param raise_events:
        """
        if isinstance(event, TransactionStartEvent):
            # Update the ACS before sharing this information with interested parties
            LOG.info('Processing transaction: %s, %s, %s', event.offset, event.workflow_id, event.command_id)
            for contract_event in event.contract_events:
                if isinstance(contract_event, ContractCreateEvent):
                    self._acs.handle_create(contract_event)
                elif isinstance(contract_event, ContractArchiveEvent):
                    self._acs.handle_archive(contract_event)

        elif isinstance(event, TransactionEndEvent):
            # Notify anything waiting on commands to complete that some of them will have
            # completed as a consequence of this transaction.
            for cmd in self._writer.inflight_commands:
                cmd.notify_read_done(event.command_id, event.time)

            LOG.debug('evt recv: party %s, BIM %r (%s events)',
                      self.party, event.command_id[0:7], len(event.contract_events))

        elif isinstance(event, ActiveContractSetEvent):
            for contract_event in event.contract_events:
                self._acs.handle_create(contract_event)

        if raise_events:
            fut = self.emit_event(event)
        else:
            fut = completed(None)

        if isinstance(event, OffsetEvent):
            self._reader.offset = event.offset

        return fut

    # endregion

    # region Write path

    def write_commands(
            self,
            commands: EventHandlerResponse,
            ignore_errors: bool = False,
            workflow_id: 'Optional[str]' = None) \
            -> Awaitable[None]:
        """
        Submit a command or list of commands.

        :param commands:
            The commands to send to the server.
        :param ignore_errors:
            Whether errors should be ignored for purposes of terminating the client. If ``True``,
            then a failure to send this command does not necessarily end the client.
        :param workflow_id:
            The workflow ID to use to tag submitted commands.
        :return:
            An ``asyncio.Future`` that is resolved right before the corresponding side effects have
            hit the event stream for this party. Synchronous errors are reported back immediately
            and not failed through this mechanism.
        """
        if workflow_id is None:
            workflow_id = uuid.uuid4().hex
        cb = CommandBuilder.coerce(commands, atomic_default=True)
        cb.defaults(workflow_id=workflow_id)

        p = _PendingCommand(cb)
        p.future.add_done_callback(lambda _: self._process_command_finished(p, ignore_errors))
        self._writer.pending_commands.put(p)
        LOG.debug('write_commands(%s)', p)
        return p.future

    def _process_command_finished(self, pending_command, ignore_errors):
        try:
            if pending_command.future.exception() is None:
                LOG.debug('Command finished: %s', pending_command)
            self._writer.inflight_commands.remove(pending_command)
        except ValueError:
            LOG.warning('Tried to remove %s even though it was already removed.', pending_command)

        if pending_command.future.exception() is not None:
            # TODO: more with this; maybe let the user respond to this
            LOG.exception('A command submission failed!',
                          exc_info=pending_command.future.exception())

    async def main_writer(self):
        """
        Main coroutine for submitting commands.
        """
        LOG.info('Writer loop for party %s is starting...', self.party)
        ledger_fut = ensure_future(self._pool.ledger())

        client = self._client_fut.result()  # type: LedgerClient
        metadata = ledger_fut.result()  # type: LedgerMetadata
        validator = ValidateSerializer(metadata.store)

        self._writer.pending_commands.start()

        # Asynchronously iterate over all pending commands that a user has (or will) send.
        # This asynchronous loop "blocks" when pending_commands is empty and is "woken up"
        # immediately when new data is added to it. The loop terminates when the pending_commands
        # ServiceQueue is stopped.
        async for p in self._writer.pending_commands:
            LOG.debug('Sending a command: %s', p)
            ledger_effective_time = metadata.time_model.get_time()
            command_payloads = []  # type: List[Tuple[_PendingCommand, Sequence[CommandPayload]]]

            self._writer.inflight_commands.append(p)
            try:
                defaults = CommandDefaults(
                    default_party=self.party,
                    default_ledger_id=metadata.ledger_id,
                    default_workflow_id=None,
                    default_application_id=self._config.application_name,
                    default_command_id=None,
                    default_ttl=timedelta(seconds=30))
                cps = p.command.build(defaults, ledger_effective_time)
                if cps:
                    commands = [replace(cp, commands=validator.serialize_commands(cp.commands))
                                for cp in cps]  # type: Sequence[CommandPayload]
                    command_payloads.append((p, commands))
                    await submit_command_async(client, p, commands)
                else:
                    # This is a "null command"; don't even bother sending to the server. Immediately
                    # resolve the future successfully and discard
                    p.future.set_result(None)
            except Exception as ex:
                LOG.exception("Tried to send a command and failed!")
                p.notify_read_fail(ex)

        LOG.info('Writer loop for party %s is winding down.', self.party)

        # After the pending command list is fully empty (and never to be filled again), wait for
        # all outstanding commands.
        await gather(*(pc.future for pc in self._writer.inflight_commands), return_exceptions=True)

        LOG.info('Writer loop for party %s is finished.', self.party)

    def writer_idle(self):
        return not bool(self._writer.pending_commands) and not self._writer.inflight_commands

    def stop_writer(self):
        """
        Prohibit future command submissions from being accepted.
        """
        self._writer.pending_commands.stop()

    # endregion


@dataclass
class _PartyClientReaderState:
    offset: Optional[str] = field(default=None)


@dataclass
class _PartyClientWriterState:
    pending_commands: 'ServiceQueue[_PendingCommand]' = field(default_factory=ServiceQueue)
    inflight_commands: 'List[_PendingCommand]' = field(default_factory=list)


class _PendingCommand:
    """
    Track the status of a set of commands in-flight.
    """

    __slots__ = ('command', 'command_ids', 'max_record_time', 'future')

    def __init__(self, command: CommandBuilder):
        self.command = safe_cast(CommandBuilder, command)
        self.command_ids = None  # type: Optional[Set[str]]
        self.max_record_time = None  # type: Optional[datetime]
        self.future = get_event_loop().create_future()

    def notify_write(self, command_ids: Collection[str], max_record_time: datetime):
        if self.max_record_time is not None:
            raise Exception('cannot send an already in-progress command')
        self.command_ids = set(command_ids)
        self.max_record_time = max_record_time

    def notify_read_done(self, command_id: str, ledger_time: Optional[datetime]):
        """
        Trigger an appropriate response given the receipt of a transaction.
        """
        if self.command_ids is not None:
            if command_id not in self.command_ids:
                return

            self.command_ids.discard(command_id)
            if not self.command_ids:
                # the command is finished
                self.future.set_result(None)
                return

        if self.max_record_time is not None and ledger_time is not None and \
                self.max_record_time < ledger_time:
            self.future.set_exception(CommandTimeoutError())

    def notify_read_fail(self, ex: Exception):
        self.future.set_exception(ex)

    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return id(self)

    def __repr__(self):
        with StringIO() as buf:
            buf.write('<_PendingCommand(command=')
            buf.write(format(self.command, 'c'))
            if self.command_ids is not None:
                buf.write(', command_ids=')
                buf.write(format(self.command_ids))
            if self.max_record_time is not None:
                buf.write(', max_record_time="')
                buf.write(format(self.max_record_time.isoformat()))
                buf.write('"')
            buf.write(', future=')
            buf.write(format(self.future))
            buf.write('>')
            return buf.getvalue()


def submit_command_async(
        client: LedgerClient, p: '_PendingCommand', commands: Sequence[CommandPayload]) \
        -> Awaitable[None]:
    """
    Submit a command asynchronously.

    The returned future is resolved when all commands have successfully completed their network
    operations, but not necessarily when the ledger has committed the results of that command to the
    ledger.

    :param client:
    :param p:
    :param commands:
    :return:
    """
    coros = []
    command_ids = []
    maximum_record_time = commands[0].maximum_record_time

    for payload in commands:
        coro = None
        command_ids.append(payload.command_id)
        try:
            coro = ensure_future(client.commands(payload))

            LOG.info('cmd submit: party %s, workflow_id %r, command_id %r, %s: %r',
                     payload.party, payload.workflow_id[0:7], payload.command_id[0:7],
                     n_things(len(payload.commands), 'commands'), reprlib.repr(payload.commands))
        except Exception as ex:
            # TODO: consider what to do in this case
            p.notify_read_fail(ex)

        if coro is not None:
            coros.append(coro)

    p.notify_write(command_ids, maximum_record_time)
    if len(coros) == 0:
        return completed(None)
    elif len(coros) == 1:
        return coros[0]
    else:
        return named_gather('SubmitCommandAsync()', *coros, return_exceptions=True)


# noinspection PyShadowingBuiltins
def wrap_as_command_submission(submit_fn, callback, filter) \
        -> Callable[[ContractId, ContractData], Awaitable[Any]]:
    """
    Normalize a callback to something that takes a single contract ID and contract data, and
    return an awaitable that is resolved when the underlying command has been fully submitted.
    """
    import inspect

    @wraps(callback)
    def implementation(*args, **kwargs):
        if filter is not None and not filter(*args, **kwargs):
            return completed(None)

        try:
            ret = callback(*args, **kwargs)
        except BaseException as exception:
            LOG.exception('The callback %r threw an exception!', callback)
            return failed(exception)

        if ret is None:
            return completed(None)
        elif isinstance(ret, (CommandBuilder, Command, list, tuple)):
            try:
                ret_fut = submit_fn(ret)
            except BaseException as exception:
                LOG.exception('The callback %r returned commands that could not be submitted! (%s)',
                              callback, ret)
                return failed(exception)
            return ret_fut
        elif inspect.isawaitable(ret):
            # the user-provided callback returned an Awaitable
            cmd_fut = ensure_future(ret)
            if cmd_fut.done():
                if cmd_fut.cancelled() or cmd_fut.exception() is not None:
                    # a cancelled or failed user-provided callback Future is the same as the
                    # command submission itself failing
                    return cmd_fut

                # functionally equivalent to the non-Awaitable case if the Awaitable has already
                # completed
                return submit_fn(cmd_fut.result())
            else:
                # create `fut`, which we'll give to the user; wait for `cmd_fut` to finish, then
                # take the result of that awaitable and try to submit a command with that result
                fut = get_event_loop().create_future()

                def cmd_future_finished(_):
                    ret = cmd_fut.result()
                    if ret is None:
                        fut.set_result(None)
                    elif isinstance(ret, (CommandBuilder, Command, list, tuple)):
                        propagate(ensure_future(submit_fn(ret)), fut)
                    elif inspect.isawaitable(ret):
                        LOG.error('A callback cannot return an Awaitable of an Awaitable')
                        raise InvalidStateError(
                            'A callback cannot return an Awaitable of an Awaitable')

                cmd_fut.add_done_callback(cmd_future_finished)

                return fut
        else:
            LOG.error('the callback %r returned a value of an unexpected type: %s', callback, ret)
            raise ValueError('unexpected return type from a callback')

    return implementation


def _find_party_config(config: 'Optional[dict]', party: Party) -> 'Optional[dict]':
    """
    Look within a config dictionary (as specified by FlatConfig) for a configuration object that is
    specific to this party.
    """
    if config is not None:
        party_config_list = config.get('parties')
        if party_config_list is not None:
            for party_config in party_config_list:
                local_party = party_config.get('party')
                if local_party == party:
                    return party_config
    return None



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains utilities required by :mod:`dazl.client.config`.
"""

import logging

from argparse import ArgumentParser
from dataclasses import dataclass, field, fields, Field
from typing import Any, Callable, FrozenSet, Optional, Sequence, Tuple

from ..model.core import Party


def config_field(
        description: str,
        param_type: 'Optional[ConfigParameterType]' = None,
        default_value: Any = None,
        short_alias: Optional[str] = None,
        deprecated_alias: Optional[str] = None,
        environment_variable: Optional[str] = None) -> Field:
    return field(default=default_value, metadata={'dazl.config': ConfigParameter(
        description=description,
        param_type=param_type,
        default_value=default_value,
        short_aliases=frozenset([short_alias]) if short_alias else frozenset(),
        deprecated_aliases=frozenset([deprecated_alias]) if deprecated_alias else frozenset(),
        environment_variable=environment_variable
    )})


def config_fields(class_or_instance: Any) -> 'Sequence[Tuple[Field, ConfigParameter]]':
    return [(fld, fld.metadata.get('dazl.config')) for fld in fields(class_or_instance)]


@dataclass(frozen=True)
class ConfigParameter:
    description: str
    param_type: 'Optional[ConfigParameterType]' = None
    default_value: Any = None
    short_aliases: FrozenSet[str] = field(default_factory=frozenset)
    alternate_keys: FrozenSet[str] = field(default_factory=frozenset)
    deprecated_aliases: FrozenSet[str] = field(default_factory=frozenset)
    environment_variable: Optional[str] = None


@dataclass(frozen=True)
class ConfigParameterType:
    metavar: Optional[str]
    value_ctor: Callable[[Any], Any]


def add_argument(
        parser: 'ArgumentParser',
        key: str,
        config_param: 'ConfigParameter') -> None:
    if config_param.param_type is not None:
        aliases = ['--' + key.replace('_', '-')]
        for alias in config_param.short_aliases:
            aliases.append('-' + alias)
        for alias in config_param.alternate_keys:
            aliases.append('--' + alias.replace('_', '-'))
        for alias in config_param.deprecated_aliases:
            aliases.append('--' + alias.replace('_', '-'))

        if config_param.param_type.value_ctor == bool:
            parser.add_argument(
                *aliases,
                help=config_param.description,
                action='store_true')
        else:
            default_value = None
            if config_param.environment_variable is not None:
                import os
                default_value = os.getenv(config_param.environment_variable)
            if not default_value:
                default_value = config_param.default_value

            parser.add_argument(
                *aliases,
                type=config_param.param_type.value_ctor,
                metavar=config_param.param_type.metavar,
                help=config_param.description,
                default=default_value)


def _parse_url(obj):
    return obj


def _parse_log_level(obj) -> int:
    if not obj:
        return logging.WARNING
    if isinstance(obj, str):
        obj = obj.upper()
    p = logging.getLevelName(obj)
    return p


def _parse_str(obj: Any) -> Optional[str]:
    return str(obj) if obj is not None else None


def _parse_block_offset(obj):
    return obj


def _parse_party_list(obj) -> 'Sequence[Party]':
    if isinstance(obj, str):
        return [Party(p) for p in obj.split(',')]
    else:
        return obj


BOOLEAN_TYPE = ConfigParameterType(None, bool)
COUNT_TYPE = ConfigParameterType('COUNT', int)
BLOCK_OFFSET_TYPE = ConfigParameterType('OFFSET', _parse_block_offset)
LOG_LEVEL_TYPE = ConfigParameterType('LOG_LEVEL', _parse_log_level)
SECONDS_TYPE = ConfigParameterType('SECONDS', float)
URL_TYPE = ConfigParameterType('URL', _parse_url)
PARTIES_TYPE = ConfigParameterType('PARTIES', _parse_party_list)
PORT_TYPE = ConfigParameterType('PORT', int)
STRING_TYPE = ConfigParameterType('STRING', _parse_str)
PATH_TYPE = ConfigParameterType('PATH', _parse_str)
VERIFY_SSL_TYPE = ConfigParameterType('VERIFY_SSL', _parse_str)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['dazl',
 'dazl._gen',
 'dazl._gen.com',
 'dazl._gen.com.digitalasset',
 'dazl._gen.com.digitalasset.ledger',
 'dazl._gen.com.digitalasset.ledger.api',
 'dazl._gen.com.digitalasset.ledger.api.v1',
 'dazl._gen.com.digitalasset.ledger.api.v1.admin',
 'dazl._gen.com.digitalasset.ledger.api.v1.testing',
 'dazl._gen.da',
 'dazl._gen.google',
 'dazl._gen.google.rpc',
 'dazl._gen.grpc',
 'dazl._gen.grpc.health',
 'dazl._gen.grpc.health.v1',
 'dazl.cli',
 'dazl.client',
 'dazl.damlast',
 'dazl.damleval',
 'dazl.damlsdk',
 'dazl.metrics',
 'dazl.model',
 'dazl.plugins',
 'dazl.plugins.capture',
 'dazl.pretty',
 'dazl.protocols',
 'dazl.protocols.admin',
 'dazl.protocols.v0',
 'dazl.protocols.v1',
 'dazl.protocols.v1.model',
 'dazl.rules',
 'dazl.server',
 'dazl.util']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp',
 'google-auth',
 'grpcio>=1.20.1',
 'oauthlib',
 'protobuf>=3.7.1',
 'pyyaml',
 'requests',
 'semver',
 'toposort']

extras_require = \
{':python_version >= "3.6.0" and python_version < "3.7.0"': ['dataclasses'],
 'prometheus': ['prometheus_client'],
 'pygments': ['pygments']}

entry_points = \
{'console_scripts': ['dazl = dazl.cli:main']}

setup_kwargs = {
    'name': 'dazl',
    'version': '5.7.0',
    'description': 'high-level Ledger API client for DAML ledgers',
    'long_description': 'dazl\n====\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](https://github.com/digital-asset/dazl-client/blob/master/LICENSE)\n<a href="https://circleci.com/gh/digital-asset/dazl-client">\n<img src="https://circleci.com/gh/digital-asset/dazl-client.svg?style=svg">\n</a>\n\nCopyright 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All Rights Reserved.\nSPDX-License-Identifier: Apache-2.0\n\n\nRich Python bindings for accessing Ledger API-based applications.\n\nRequirements\n------------\n* Python 3.6+\n* [Pipenv](https://pipenv.readthedocs.io/en/latest/)\n* Although not strictly required for building, you\'ll probably want the [DAML SDK](https://www.daml.com)\n\nExamples\n--------\n\nAll of the examples below assume you imported `dazl`.\n\nConnect to the ledger and submit a single command:\n\n```py\nwith dazl.simple_client(\'http://localhost:7600\', \'Alice\') as client:\n    client.submit_create(\'Alice\', \'My.Template\', { someField: \'someText\' })\n```\n\nConnect to the ledger as a single party, print all contracts, and close:\n\n```py\nwith dazl.simple_client(\'http://localhost:7600\', \'Alice\') as client:\n    # wait for the ACS to be fully read\n    client.ready()\n    contract_dict = client.find_active(\'*\')\nprint(contract_dict)\n```\n\nConnect to the ledger as multiple parties:\n\n```py\nnetwork = dazl.Network()\nnetwork.set_config(url=\'http://localhost:7600\')\n\nalice = network.simple_party(\'Alice\')\nbob = network.simple_party(\'Bob\')\n\n@alice.ledger_ready()\ndef set_up(event):\n    currency_cid, _ = await event.acs_find_one(\'My.Currency\', {"currency": "USD"})\n    return dazl.create(\'SomethingOf.Value\', {\n        \'amount\': 100,\n        \'currency\': currency_cid,\n        \'from\': \'Accept\',\n        \'to\': \'Bob\' })\n\n@bob.ledger_created(\'SomethingOf.Value\')\ndef on_something_of_value(event):\n    return dazl.exercise(event.cid, \'Accept\', { \'message\': \'Thanks!\' })\n\nnetwork.start()\n```\n\n\nBuilding locally\n----------------\n```sh\ncd python && pipenv run package\n```\n\nTests\n-----\n\n```sh\ncd python && pipenv run test\n```\n',
    'author': 'Davin K. Tanabe',
    'author_email': 'davin.tanabe@digitalasset.com',
    'url': 'https://github.com/digital-asset/dazl-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from functools import reduce
from operator import add
from threading import RLock
from types import MappingProxyType

from dataclasses import dataclass
from typing import Any, Collection, Dict, Generic, List, Mapping, Optional, Sequence, TypeVar, \
    Union, TYPE_CHECKING

from .types import Template, TemplateChoice, Type, TypeReference, UnresolvedTypeReference, \
    TypeAdjective, ConcreteType, ValueReference
from ..util.typing import safe_cast, safe_dict_cast

if TYPE_CHECKING:
    from ..damlast.daml_lf_1 import Archive, Expr, Package

T = TypeVar('T')


class PackageStoreBuilder:
    """
    Convenience class for building up a :class:`PackageStore`.
    """

    def __init__(self):
        self._archives = list()  # type: List[Archive]
        self._value_types = dict()  # type: Dict[ValueReference, Expr]
        self._data_types = dict()  # type: Dict[TypeReference, Type]
        self._templates = dict()  # type: Dict[TypeReference, Template]

    def add_archive(self, archive: 'Archive') -> None:
        self._archives.append(archive)

    def add_type(self, name: TypeReference, data_type: Type):
        safe_cast(TypeReference, name)
        safe_cast(Type, data_type)
        self._data_types[name] = data_type

    def add_value(self, name: ValueReference, value: 'Expr'):
        self._value_types[name] = value

    def add_template(self, template: Template):
        self._templates[template.data_type.name] = safe_cast(Template, template)

    def get_type(self, name: TypeReference) -> Optional[Type]:
        return self._data_types.get(name)

    def build(self) -> 'PackageStore':
        return PackageStore(self._archives, self._value_types, self._data_types, self._templates)


class PackageStore:
    """
    A thread-safe store of type information.
    """

    @classmethod
    def empty(cls):
        """
        Create an empty store.
        """
        return cls([], {}, {}, {})

    def __init__(
            self,
            archives: 'Collection[Archive]',
            value_types: 'Dict[ValueReference, Expr]',
            data_types: 'Dict[TypeReference, Type]',
            templates: 'Dict[TypeReference, Template]'):
        self._lock = RLock()
        self._archives = list(archives)
        self._cache = PackageStoreCache(EMPTY_TYPE_CACHE, EMPTY_TYPE_CACHE, EMPTY_TYPE_CACHE)
        self._value_types = value_types  # safe_dict_cast(ValueReference, Expr, value_types)
        self._data_types = safe_dict_cast(TypeReference, Type, data_types)
        self._templates = safe_dict_cast(TypeReference, Template, templates)

    def archives(self) -> 'Collection[Archive]':
        """
        Return a copy of the collection of the set of loaded :class:`Archive`s.
        """
        with self._lock:
            return list(self._archives)

    def packages(self) -> 'Collection[Package]':
        """
        Return a copy of the collection of the set of loaded :class:`Package`s.
        """
        with self._lock:
            return [a.package for a in self._archives]

    def package_ids(self) -> 'Collection[str]':
        """
        Return a copy of the collection of the set of loaded :class:`Package`s.
        """
        with self._lock:
            return [a.hash for a in self._archives]

    def register_all(self, other_store: 'PackageStore') -> 'PackageStore':
        """
        Register all types declared in the other :class:`PackageStore`.

        :param other_store: A package store to copy types, templates, and choices from.
        :return: A reference to this object.
        """
        if self is not other_store:
            with self._lock:
                self._archives.extend(other_store._archives)
                self._value_types.update(other_store._value_types)
                self._data_types.update(other_store._data_types)
                self._templates.update(other_store._templates)
                self._cache = PackageStoreCache(
                    TypeCache.build(self._value_types),
                    TypeCache.build(self._data_types),
                    TypeCache.build(self._templates))
        return self

    def resolve_value_reference(self, value_ref: ValueReference) -> 'Expr':
        with self._lock:
            return self._value_types[value_ref]

    def resolve_type_reference(self, template_ref: TypeReference) -> Type:
        """
        Resolve a type based on its reference.

        :param template_ref:
            A :class:`TypeReference` that refers to a type.
        :return:
            The :class:`Type` that is referred to by this type reference.
        :raise KeyError:
            If the :class:`TypeReference` does not have a corresponding value in this store.
        """
        safe_cast(TypeReference, template_ref)

        with self._lock:
            return self._data_types[template_ref]

    def find_types(self, adjective: TypeAdjective = TypeAdjective.ANY) \
            -> Mapping[TypeReference, ConcreteType]:
        with self._lock:
            if adjective == TypeAdjective.ANY:
                return MappingProxyType(dict(self._data_types))

            return {tt: data_type
                    for tt, data_type in self._data_types.items()
                    if isinstance(data_type, ConcreteType)
                    and data_type.adjective & adjective != TypeAdjective.NONE}

    def resolve_template(self, template: Union[None, str, TypeReference, UnresolvedTypeReference]) \
            -> Collection[Template]:
        """
        Return a collection of :class:`Template` instances that match the specified template name.

        Some special parameters:
         * If ``"*"`` is passed in, all templates are returned.
         * If ``None`` is passed in, an empty collection is returned.

        :param template:
            A template name or a :class:`TypeReference`.
        :return:
            A collection of matching templates, or an empty collection if none match. This method
            never returns ``None``.
        """
        from .lookup import validate_template

        package_id, template_name = validate_template(template)
        return self._cache.templates.lookup(package_id, template_name)

    def resolve_template_type(
            self, template: 'Union[None, str, TypeReference, UnresolvedTypeReference]') \
            -> Dict[TypeReference, Type]:
        """
        Return a collection of types that match for the template.

        :param template:
            A template name or a :class:`TypeReference`.
        :return:
            A dictionary of possible matches, or empty if there are no matches. This method never
            returns ``None``.
        """
        return {template.data_type.name: template.data_type
                for template in self.resolve_template(template)}

    def resolve_choice(self, template: Any, choice: str) -> Dict[TypeReference, TemplateChoice]:
        """
        Return all possible choices for the combination of template identifier and choice name.
        If template is ``'*'`` or ``None``, all choices with the specified name are returned.
        """
        matches = dict()
        for t in self.resolve_template(template):
            for c in t.choices:
                if c.name == choice:
                    matches[t.data_type.name] = c
        return matches


class PackageProvider:
    """
    Interface to an object that can provide package information.
    """

    def get_package_ids(self) -> 'Sequence[str]':
        """
        Return the current universe of package IDs.
        """
        raise NotImplementedError

    def fetch_package(self, package_id: str) -> bytes:
        """
        Retrieve the bytes that correspond to a package.
        """
        raise NotImplementedError

    def get_all_packages(self) -> 'Mapping[str, bytes]':
        return {pkg_id: self.fetch_package(pkg_id) for pkg_id in self.get_package_ids()}


class MemoryPackageProvider(PackageProvider):
    def __init__(self, mapping: 'Mapping[str, bytes]'):
        self.mapping = mapping

    def get_package_ids(self) -> 'Sequence[str]':
        return list(self.mapping.keys())

    def fetch_package(self, package_id: str) -> bytes:
        return self.mapping.get(package_id)


@dataclass(frozen=True)
class PackageStoreCache:
    value_types: 'TypeCache[Type]'
    data_types: 'TypeCache[Type]'
    templates: 'TypeCache[Template]'


@dataclass(frozen=True)
class TypeCache(Generic[T]):
    """
    Immutable cache of type information.

    Instance attributes:

    .. attribute:: TypeCache.by_package_lookup

        A mapping from package ID strings to another mapping, where keys are fully-qualified type
        names for templates to a collection of matching Template instances. These collections are
        never non-empty; they will only ever contain multiple entries if there is a collision in
        dot-separated names. None of these maps contain ``'*'`` as keys; wildcards mut be
        implemented by the caller.

    .. attribute:: TypeCache.by_name_lookup

        A mapping from fully-qualified type names to a sub-mapping, where the sub-mapping keys are
        package IDs and the value are collections of Template. None of these maps contain ``'*'``
        as keys; wildcards mut be implemented by the caller.
    """

    everything: Collection['T']
    by_package_lookup: 'Mapping[str, Mapping[str, Collection[T]]]'
    by_name_lookup: 'Mapping[str, Mapping[str, Collection[T]]]'

    @classmethod
    def build(cls, objects: 'Mapping[TypeReference, T]') -> 'TypeCache[T]':
        everything = tuple(objects)
        by_package_lookup = defaultdict(lambda: defaultdict(list))
        by_name_lookup = defaultdict(lambda: defaultdict(list))

        for k, v in objects.items():
            for package_id in (k.module.package_id, '*'):
                for valid_name in (k.full_name, k.full_name_unambiguous, '*'):
                    by_package_lookup[package_id][valid_name].append(v)
                    by_name_lookup[valid_name][package_id].append(v)

        return TypeCache(
            everything,
            _immutable_mmc(by_package_lookup),
            _immutable_mmc(by_name_lookup))

    def lookup(self, package_id: str, type_name: str) -> Collection[T]:
        """
        Look up items based on package ID and type name. The values cannot be ``None``, but they
        can be the special-meaning ``'*'`` value.
        """
        safe_cast(str, package_id)
        safe_cast(str, type_name)

        if package_id == '*':
            candidates = reduce(add, self.by_name_lookup.get(type_name, {}).values(), ())
        elif type_name == '*':
            candidates = reduce(add, self.by_package_lookup.get(package_id, {}).values(), ())
        else:
            candidates = self.by_package_lookup.get(package_id, {}).get(type_name, ())

        return {t.data_type.name.full_name_unambiguous: t for t in candidates}.values()


def _immutable_mmc(mapping: 'Mapping[str, Mapping[str, Collection[T]]]') -> \
        'Mapping[str, Mapping[str, Collection[T]]]':
    """
    Create an immutable copy of :class:`TemplateStoreCache` data structures.
    """
    return MappingProxyType({
        k1: MappingProxyType({k2: tuple(v) for k2, v in v1.items()})
        for k1, v1 in mapping.items()
    })


EMPTY_MAPPING = MappingProxyType({})
EMPTY_TYPE_CACHE = TypeCache((), EMPTY_MAPPING, EMPTY_MAPPING)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Union


def pathify(path: Union[str, Path]) -> Path:
    """
    Convert an object that could be a :class:`Path` into a :class:`Path`.

    :param path:
    :return:
    """
    if isinstance(path, Path):
        return path
    elif isinstance(path, str):
        return Path(path)
    else:
        raise ValueError(f'path must be a str or Path (got {path!r} instead)')



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Formatting module for outputting captures.
"""

from typing import Iterable

from ...protocols.v0.json_ser_command import LedgerJSONEncoder
from ...util.tools import boundary_iter
from .model_capture import LedgerCapture


def format_error(error: str) -> Iterable[str]:
    encode = LedgerJSONEncoder().encode
    return encode(dict(errors=[error]))


def format_entries(capture: LedgerCapture, parties=None, entries=None, **kwargs) -> Iterable[str]:
    """
    Return a list of strings that, when taken together, constitute a JSON document that lists
    all of the entries of the ledger.

    :param capture:
        The contents to render.
    :param parties:
        The list of of parties.
    :param entries:
        The entries to render.
    :param kwargs:
        Additional parameters. Because all formatters have different arguments, these extra
        unknown parameters are ignored.
    """
    encode = LedgerJSONEncoder().encode

    yield "{{ \"parties\": {},".format(encode(parties))
    yield "  \"contracts\": ["

    for is_last, entry in boundary_iter(entries):
        party_list = sorted(entry.parties.keys())

        yield '  {{ "contract_id": "{}",'.format(entry.contract_id)
        yield '    "template": {},'.format(encode(entry.template_id.full_name))
        yield '    "parties": {},'.format(encode(party_list))
        yield '    "arguments": {} }}{}'.format(encode(entry.contract_args), '' if is_last else ',')

    yield "  ] }"



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Define a few DAML values and functions that have been inlined in Python for efficiency and
simplicity.
"""

from typing import Any, Optional, Sequence

from ._builtins_meta import BuiltinTable, Builtin
from .daml_lf_1 import BuiltinFunction, Expr, PrimLit, Type
from .util import list_type, var, STRING

builtins = BuiltinTable()


@builtins.add
class Concat(Builtin):
    name = 'DA.Internal.Prelude:concat'
    args = list_type(list_type(var('x'))),

    def evaluate(self, _: 'Sequence[Type]', args: 'Sequence[Any]') -> 'Any':
        xxs = args[0]

        result = []
        for xs in xxs:
            result.extend(xs)
        return result

    def simplify(self, _: 'Sequence[Type]', args: 'Sequence[Expr]') -> 'Optional[Expr]':
        xxs = args[0]

        # if there is a single-element list as an arg to concat, simply remove the function call
        # and fully unwrap
        if xxs.cons is not None and xxs.cons.tail.nil is not None and len(xxs.cons.front) == 1:
            return xxs.cons.front[0]
        if xxs.nil is not None:
            return Expr(nil=Expr.Nil(xxs.nil.type.prim.args[0]))


@builtins.add
class Show(Builtin):
    name = 'GHC.Show:show'
    args = (STRING, STRING)

    def evaluate(self, _: 'Sequence[Type]', args: 'Sequence[Any]') -> 'Any':
        return str(args[1])

    def simplify(self, _: 'Sequence[Type]', args: 'Sequence[Expr]') -> 'Optional[Expr]':
        return None


@builtins.add
class AppendText(Builtin):
    builtin = BuiltinFunction.APPEND_TEXT
    args = (STRING, STRING)

    def evaluate(self, _: 'Sequence[Type]', args: 'Sequence[Any]') -> 'Any':
        return args[0] + args[1]

    def simplify(self, _: 'Sequence[Type]', args: 'Sequence[Expr]') -> 'Expr':
        if args[0].prim_lit is not None and args[1].prim_lit is not None:
            return Expr(prim_lit=PrimLit(text=args[0].prim_lit.text + args[1].prim_lit.text))






# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


DEFAULT_FORMATTER_NAME = 'pretty'


def get_formatter(fmt):
    """
    Return a formatter of the specified name.
    """
    if fmt is None:
        fmt = DEFAULT_FORMATTER_NAME

    if fmt == 'json':
        from . import fmt_json
        return fmt_json
    elif fmt == 'pretty':
        from . import fmt_pretty
        return fmt_pretty
    else:
        raise ValueError('unknown format: {}'.format(fmt))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Utilities for working with the built-in `logging` module.
"""
import logging
import threading
from typing import cast

from ..util.io import StdoutStreamWrapper

VERBOSE = 5

LOG_LOCK = threading.Lock()
_default_logger_init = False


def setup_default_logger(level=logging.INFO):
    """
    Sets up a default logger with sensible defaults.
    """
    root = logging.getLogger()
    global _default_logger_init
    if not _default_logger_init:
        _default_logger_init = True
    else:
        root.warning('setup_default_logger being called more than once!')
        return

    logging.captureWarnings(True)

    root.setLevel(level)
    stream_handler = logging.StreamHandler(StdoutStreamWrapper())
    formatter = logging.Formatter("[%(levelname)7s] %(asctime)s | %(name)-7s | %(message)s")
    stream_handler.setFormatter(formatter)
    root.addHandler(stream_handler)


def log_verbose(logger: 'logging.Logger', msg, *args, **kwargs) -> None:
    return logger.log(5, msg, *args, **kwargs)


class _LoggerWithVerbose:
    def verbose(self, msg, *args, **kwargs) -> None:
        logger = cast(logging.Logger, self)
        return log_verbose(logger, msg, *args, **kwargs)


class LoggerWithVerbose(logging.Logger, _LoggerWithVerbose):
    pass


logging.addLevelName(VERBOSE, 'VERBOSE')



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/admin/party_management_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/admin/party_management_service.proto',
  package='com.digitalasset.ledger.api.v1.admin',
  syntax='proto3',
  serialized_options=_b('\n$com.digitalasset.ledger.api.v1.adminB PartyManagementServiceOuterClass'),
  serialized_pb=_b('\nCcom/digitalasset/ledger/api/v1/admin/party_management_service.proto\x12$com.digitalasset.ledger.api.v1.admin\"\x19\n\x17GetParticipantIdRequest\"2\n\x18GetParticipantIdResponse\x12\x16\n\x0eparticipant_id\x18\x01 \x01(\t\"\x19\n\x17ListKnownPartiesRequest\"e\n\x18ListKnownPartiesResponse\x12I\n\rparty_details\x18\x01 \x03(\x0b\x32\x32.com.digitalasset.ledger.api.v1.admin.PartyDetails\"E\n\x0cPartyDetails\x12\r\n\x05party\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x10\n\x08is_local\x18\x03 \x01(\x08\"C\n\x14\x41llocatePartyRequest\x12\x15\n\rparty_id_hint\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\"b\n\x15\x41llocatePartyResponse\x12I\n\rparty_details\x18\x01 \x01(\x0b\x32\x32.com.digitalasset.ledger.api.v1.admin.PartyDetails2\xcb\x03\n\x16PartyManagementService\x12\x91\x01\n\x10GetParticipantId\x12=.com.digitalasset.ledger.api.v1.admin.GetParticipantIdRequest\x1a>.com.digitalasset.ledger.api.v1.admin.GetParticipantIdResponse\x12\x91\x01\n\x10ListKnownParties\x12=.com.digitalasset.ledger.api.v1.admin.ListKnownPartiesRequest\x1a>.com.digitalasset.ledger.api.v1.admin.ListKnownPartiesResponse\x12\x88\x01\n\rAllocateParty\x12:.com.digitalasset.ledger.api.v1.admin.AllocatePartyRequest\x1a;.com.digitalasset.ledger.api.v1.admin.AllocatePartyResponseBH\n$com.digitalasset.ledger.api.v1.adminB PartyManagementServiceOuterClassb\x06proto3')
)




_GETPARTICIPANTIDREQUEST = _descriptor.Descriptor(
  name='GetParticipantIdRequest',
  full_name='com.digitalasset.ledger.api.v1.admin.GetParticipantIdRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=109,
  serialized_end=134,
)


_GETPARTICIPANTIDRESPONSE = _descriptor.Descriptor(
  name='GetParticipantIdResponse',
  full_name='com.digitalasset.ledger.api.v1.admin.GetParticipantIdResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='participant_id', full_name='com.digitalasset.ledger.api.v1.admin.GetParticipantIdResponse.participant_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=136,
  serialized_end=186,
)


_LISTKNOWNPARTIESREQUEST = _descriptor.Descriptor(
  name='ListKnownPartiesRequest',
  full_name='com.digitalasset.ledger.api.v1.admin.ListKnownPartiesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=188,
  serialized_end=213,
)


_LISTKNOWNPARTIESRESPONSE = _descriptor.Descriptor(
  name='ListKnownPartiesResponse',
  full_name='com.digitalasset.ledger.api.v1.admin.ListKnownPartiesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_details', full_name='com.digitalasset.ledger.api.v1.admin.ListKnownPartiesResponse.party_details', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=215,
  serialized_end=316,
)


_PARTYDETAILS = _descriptor.Descriptor(
  name='PartyDetails',
  full_name='com.digitalasset.ledger.api.v1.admin.PartyDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='party', full_name='com.digitalasset.ledger.api.v1.admin.PartyDetails.party', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='com.digitalasset.ledger.api.v1.admin.PartyDetails.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_local', full_name='com.digitalasset.ledger.api.v1.admin.PartyDetails.is_local', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=318,
  serialized_end=387,
)


_ALLOCATEPARTYREQUEST = _descriptor.Descriptor(
  name='AllocatePartyRequest',
  full_name='com.digitalasset.ledger.api.v1.admin.AllocatePartyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_id_hint', full_name='com.digitalasset.ledger.api.v1.admin.AllocatePartyRequest.party_id_hint', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='com.digitalasset.ledger.api.v1.admin.AllocatePartyRequest.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=389,
  serialized_end=456,
)


_ALLOCATEPARTYRESPONSE = _descriptor.Descriptor(
  name='AllocatePartyResponse',
  full_name='com.digitalasset.ledger.api.v1.admin.AllocatePartyResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='party_details', full_name='com.digitalasset.ledger.api.v1.admin.AllocatePartyResponse.party_details', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=458,
  serialized_end=556,
)

_LISTKNOWNPARTIESRESPONSE.fields_by_name['party_details'].message_type = _PARTYDETAILS
_ALLOCATEPARTYRESPONSE.fields_by_name['party_details'].message_type = _PARTYDETAILS
DESCRIPTOR.message_types_by_name['GetParticipantIdRequest'] = _GETPARTICIPANTIDREQUEST
DESCRIPTOR.message_types_by_name['GetParticipantIdResponse'] = _GETPARTICIPANTIDRESPONSE
DESCRIPTOR.message_types_by_name['ListKnownPartiesRequest'] = _LISTKNOWNPARTIESREQUEST
DESCRIPTOR.message_types_by_name['ListKnownPartiesResponse'] = _LISTKNOWNPARTIESRESPONSE
DESCRIPTOR.message_types_by_name['PartyDetails'] = _PARTYDETAILS
DESCRIPTOR.message_types_by_name['AllocatePartyRequest'] = _ALLOCATEPARTYREQUEST
DESCRIPTOR.message_types_by_name['AllocatePartyResponse'] = _ALLOCATEPARTYRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetParticipantIdRequest = _reflection.GeneratedProtocolMessageType('GetParticipantIdRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETPARTICIPANTIDREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.GetParticipantIdRequest)
  ))
_sym_db.RegisterMessage(GetParticipantIdRequest)

GetParticipantIdResponse = _reflection.GeneratedProtocolMessageType('GetParticipantIdResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETPARTICIPANTIDRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.GetParticipantIdResponse)
  ))
_sym_db.RegisterMessage(GetParticipantIdResponse)

ListKnownPartiesRequest = _reflection.GeneratedProtocolMessageType('ListKnownPartiesRequest', (_message.Message,), dict(
  DESCRIPTOR = _LISTKNOWNPARTIESREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.ListKnownPartiesRequest)
  ))
_sym_db.RegisterMessage(ListKnownPartiesRequest)

ListKnownPartiesResponse = _reflection.GeneratedProtocolMessageType('ListKnownPartiesResponse', (_message.Message,), dict(
  DESCRIPTOR = _LISTKNOWNPARTIESRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.ListKnownPartiesResponse)
  ))
_sym_db.RegisterMessage(ListKnownPartiesResponse)

PartyDetails = _reflection.GeneratedProtocolMessageType('PartyDetails', (_message.Message,), dict(
  DESCRIPTOR = _PARTYDETAILS,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.PartyDetails)
  ))
_sym_db.RegisterMessage(PartyDetails)

AllocatePartyRequest = _reflection.GeneratedProtocolMessageType('AllocatePartyRequest', (_message.Message,), dict(
  DESCRIPTOR = _ALLOCATEPARTYREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.AllocatePartyRequest)
  ))
_sym_db.RegisterMessage(AllocatePartyRequest)

AllocatePartyResponse = _reflection.GeneratedProtocolMessageType('AllocatePartyResponse', (_message.Message,), dict(
  DESCRIPTOR = _ALLOCATEPARTYRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.admin.party_management_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.admin.AllocatePartyResponse)
  ))
_sym_db.RegisterMessage(AllocatePartyResponse)


DESCRIPTOR._options = None

_PARTYMANAGEMENTSERVICE = _descriptor.ServiceDescriptor(
  name='PartyManagementService',
  full_name='com.digitalasset.ledger.api.v1.admin.PartyManagementService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=559,
  serialized_end=1018,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetParticipantId',
    full_name='com.digitalasset.ledger.api.v1.admin.PartyManagementService.GetParticipantId',
    index=0,
    containing_service=None,
    input_type=_GETPARTICIPANTIDREQUEST,
    output_type=_GETPARTICIPANTIDRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='ListKnownParties',
    full_name='com.digitalasset.ledger.api.v1.admin.PartyManagementService.ListKnownParties',
    index=1,
    containing_service=None,
    input_type=_LISTKNOWNPARTIESREQUEST,
    output_type=_LISTKNOWNPARTIESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='AllocateParty',
    full_name='com.digitalasset.ledger.api.v1.admin.PartyManagementService.AllocateParty',
    index=2,
    containing_service=None,
    input_type=_ALLOCATEPARTYREQUEST,
    output_type=_ALLOCATEPARTYRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PARTYMANAGEMENTSERVICE)

DESCRIPTOR.services_by_name['PartyManagementService'] = _PARTYMANAGEMENTSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from types import MappingProxyType
from typing import Collection, Mapping, Optional

from ..model.types import ValueReference
from ..model.types_store import PackageStore
from .daml_lf_1 import Expr, Type


class EvaluationScope:
    def __init__(
            self,
            store: 'PackageStore',
            bindings: 'Mapping[str, Expr]',
            blocked_value_refs: 'Collection[ValueReference]' = (),
            depth = 0):
        self.store = store
        self.depth = depth
        self._vars = MappingProxyType(dict(bindings))
        self._blocked_value_refs = frozenset(blocked_value_refs)

    def without(self, var_name: str) -> 'EvaluationScope':
        """
        Return a child scope with the specified variable name _unbound_.

        :param var_name:
        :return:
        """
        new_bindings = dict(self._vars)
        del new_bindings[var_name]
        return EvaluationScope(self.store, new_bindings, self._blocked_value_refs, self.depth + 1)

    def without_value(self, value_ref) -> 'EvaluationScope':
        """
        Return a child scope where the value corresponding to the specified value reference will
        NOT be returned. (This is used to handle breaking otherwise recursive calls.)

        :param value_ref: The value to withhold.
        """
        values = set(self._blocked_value_refs)
        values.discard(value_ref)

        return EvaluationScope(self.store, self._vars, frozenset(values), self.depth + 1)

    def child_scope(self, new_bindings: 'Mapping[str, Expr]') -> 'EvaluationScope':
        bindings = dict(self._vars)
        bindings.update(new_bindings)
        return EvaluationScope(self.store, bindings, self._blocked_value_refs, self.depth + 1)

    def get_bound(self, var_name) -> 'Expr':
        return self._vars[var_name]

    def resolve_type(self, type_ref) -> 'Type':
        pass

    def resolve_value(self, value_ref) -> 'Optional[Expr]':
        if value_ref in self._blocked_value_refs:
            return None
        else:
            return self.store.resolve_value_reference(value_ref)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/digitalasset/ledger/api/v1/transaction_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import ledger_offset_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_ledger__offset__pb2
from . import transaction_filter_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2
from . import transaction_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__pb2
from . import trace_context_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='com/digitalasset/ledger/api/v1/transaction_service.proto',
  package='com.digitalasset.ledger.api.v1',
  syntax='proto3',
  serialized_options=_b('\n\036com.digitalasset.ledger.api.v1B\034TransactionServiceOuterClass'),
  serialized_pb=_b('\n8com/digitalasset/ledger/api/v1/transaction_service.proto\x12\x1e\x63om.digitalasset.ledger.api.v1\x1a\x32\x63om/digitalasset/ledger/api/v1/ledger_offset.proto\x1a\x37\x63om/digitalasset/ledger/api/v1/transaction_filter.proto\x1a\x30\x63om/digitalasset/ledger/api/v1/transaction.proto\x1a\x32\x63om/digitalasset/ledger/api/v1/trace_context.proto\"\xbd\x02\n\x16GetTransactionsRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12;\n\x05\x62\x65gin\x18\x02 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.LedgerOffset\x12\x39\n\x03\x65nd\x18\x03 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.LedgerOffset\x12\x41\n\x06\x66ilter\x18\x04 \x01(\x0b\x32\x31.com.digitalasset.ledger.api.v1.TransactionFilter\x12\x0f\n\x07verbose\x18\x05 \x01(\x08\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"\\\n\x17GetTransactionsResponse\x12\x41\n\x0ctransactions\x18\x01 \x03(\x0b\x32+.com.digitalasset.ledger.api.v1.Transaction\"d\n\x1bGetTransactionTreesResponse\x12\x45\n\x0ctransactions\x18\x01 \x03(\x0b\x32/.com.digitalasset.ledger.api.v1.TransactionTree\"\xa7\x01\n\x1eGetTransactionByEventIdRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x10\n\x08\x65vent_id\x18\x02 \x01(\t\x12\x1a\n\x12requesting_parties\x18\x03 \x03(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"\xa8\x01\n\x19GetTransactionByIdRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x16\n\x0etransaction_id\x18\x02 \x01(\t\x12\x1a\n\x12requesting_parties\x18\x03 \x03(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"^\n\x16GetTransactionResponse\x12\x44\n\x0btransaction\x18\x01 \x01(\x0b\x32/.com.digitalasset.ledger.api.v1.TransactionTree\"^\n\x1aGetFlatTransactionResponse\x12@\n\x0btransaction\x18\x01 \x01(\x0b\x32+.com.digitalasset.ledger.api.v1.Transaction\"n\n\x13GetLedgerEndRequest\x12\x11\n\tledger_id\x18\x01 \x01(\t\x12\x44\n\rtrace_context\x18\xe8\x07 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.TraceContext\"T\n\x14GetLedgerEndResponse\x12<\n\x06offset\x18\x01 \x01(\x0b\x32,.com.digitalasset.ledger.api.v1.LedgerOffset2\xf1\x07\n\x12TransactionService\x12\x84\x01\n\x0fGetTransactions\x12\x36.com.digitalasset.ledger.api.v1.GetTransactionsRequest\x1a\x37.com.digitalasset.ledger.api.v1.GetTransactionsResponse0\x01\x12\x8c\x01\n\x13GetTransactionTrees\x12\x36.com.digitalasset.ledger.api.v1.GetTransactionsRequest\x1a;.com.digitalasset.ledger.api.v1.GetTransactionTreesResponse0\x01\x12\x91\x01\n\x17GetTransactionByEventId\x12>.com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest\x1a\x36.com.digitalasset.ledger.api.v1.GetTransactionResponse\x12\x87\x01\n\x12GetTransactionById\x12\x39.com.digitalasset.ledger.api.v1.GetTransactionByIdRequest\x1a\x36.com.digitalasset.ledger.api.v1.GetTransactionResponse\x12\x99\x01\n\x1bGetFlatTransactionByEventId\x12>.com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest\x1a:.com.digitalasset.ledger.api.v1.GetFlatTransactionResponse\x12\x8f\x01\n\x16GetFlatTransactionById\x12\x39.com.digitalasset.ledger.api.v1.GetTransactionByIdRequest\x1a:.com.digitalasset.ledger.api.v1.GetFlatTransactionResponse\x12y\n\x0cGetLedgerEnd\x12\x33.com.digitalasset.ledger.api.v1.GetLedgerEndRequest\x1a\x34.com.digitalasset.ledger.api.v1.GetLedgerEndResponseB>\n\x1e\x63om.digitalasset.ledger.api.v1B\x1cTransactionServiceOuterClassb\x06proto3')
  ,
  dependencies=[com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_ledger__offset__pb2.DESCRIPTOR,com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2.DESCRIPTOR,com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__pb2.DESCRIPTOR,com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2.DESCRIPTOR,])




_GETTRANSACTIONSREQUEST = _descriptor.Descriptor(
  name='GetTransactionsRequest',
  full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='begin', full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest.begin', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='end', full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest.end', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='filter', full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest.filter', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='verbose', full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest.verbose', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetTransactionsRequest.trace_context', index=5,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=304,
  serialized_end=621,
)


_GETTRANSACTIONSRESPONSE = _descriptor.Descriptor(
  name='GetTransactionsResponse',
  full_name='com.digitalasset.ledger.api.v1.GetTransactionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transactions', full_name='com.digitalasset.ledger.api.v1.GetTransactionsResponse.transactions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=623,
  serialized_end=715,
)


_GETTRANSACTIONTREESRESPONSE = _descriptor.Descriptor(
  name='GetTransactionTreesResponse',
  full_name='com.digitalasset.ledger.api.v1.GetTransactionTreesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transactions', full_name='com.digitalasset.ledger.api.v1.GetTransactionTreesResponse.transactions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=717,
  serialized_end=817,
)


_GETTRANSACTIONBYEVENTIDREQUEST = _descriptor.Descriptor(
  name='GetTransactionByEventIdRequest',
  full_name='com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='event_id', full_name='com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest.event_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requesting_parties', full_name='com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest.requesting_parties', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest.trace_context', index=3,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=820,
  serialized_end=987,
)


_GETTRANSACTIONBYIDREQUEST = _descriptor.Descriptor(
  name='GetTransactionByIdRequest',
  full_name='com.digitalasset.ledger.api.v1.GetTransactionByIdRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetTransactionByIdRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='transaction_id', full_name='com.digitalasset.ledger.api.v1.GetTransactionByIdRequest.transaction_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='requesting_parties', full_name='com.digitalasset.ledger.api.v1.GetTransactionByIdRequest.requesting_parties', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetTransactionByIdRequest.trace_context', index=3,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=990,
  serialized_end=1158,
)


_GETTRANSACTIONRESPONSE = _descriptor.Descriptor(
  name='GetTransactionResponse',
  full_name='com.digitalasset.ledger.api.v1.GetTransactionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction', full_name='com.digitalasset.ledger.api.v1.GetTransactionResponse.transaction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1160,
  serialized_end=1254,
)


_GETFLATTRANSACTIONRESPONSE = _descriptor.Descriptor(
  name='GetFlatTransactionResponse',
  full_name='com.digitalasset.ledger.api.v1.GetFlatTransactionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='transaction', full_name='com.digitalasset.ledger.api.v1.GetFlatTransactionResponse.transaction', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1256,
  serialized_end=1350,
)


_GETLEDGERENDREQUEST = _descriptor.Descriptor(
  name='GetLedgerEndRequest',
  full_name='com.digitalasset.ledger.api.v1.GetLedgerEndRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ledger_id', full_name='com.digitalasset.ledger.api.v1.GetLedgerEndRequest.ledger_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trace_context', full_name='com.digitalasset.ledger.api.v1.GetLedgerEndRequest.trace_context', index=1,
      number=1000, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1352,
  serialized_end=1462,
)


_GETLEDGERENDRESPONSE = _descriptor.Descriptor(
  name='GetLedgerEndResponse',
  full_name='com.digitalasset.ledger.api.v1.GetLedgerEndResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='com.digitalasset.ledger.api.v1.GetLedgerEndResponse.offset', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1464,
  serialized_end=1548,
)

_GETTRANSACTIONSREQUEST.fields_by_name['begin'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_ledger__offset__pb2._LEDGEROFFSET
_GETTRANSACTIONSREQUEST.fields_by_name['end'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_ledger__offset__pb2._LEDGEROFFSET
_GETTRANSACTIONSREQUEST.fields_by_name['filter'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__filter__pb2._TRANSACTIONFILTER
_GETTRANSACTIONSREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETTRANSACTIONSRESPONSE.fields_by_name['transactions'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__pb2._TRANSACTION
_GETTRANSACTIONTREESRESPONSE.fields_by_name['transactions'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__pb2._TRANSACTIONTREE
_GETTRANSACTIONBYEVENTIDREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETTRANSACTIONBYIDREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETTRANSACTIONRESPONSE.fields_by_name['transaction'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__pb2._TRANSACTIONTREE
_GETFLATTRANSACTIONRESPONSE.fields_by_name['transaction'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_transaction__pb2._TRANSACTION
_GETLEDGERENDREQUEST.fields_by_name['trace_context'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_trace__context__pb2._TRACECONTEXT
_GETLEDGERENDRESPONSE.fields_by_name['offset'].message_type = com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_ledger__offset__pb2._LEDGEROFFSET
DESCRIPTOR.message_types_by_name['GetTransactionsRequest'] = _GETTRANSACTIONSREQUEST
DESCRIPTOR.message_types_by_name['GetTransactionsResponse'] = _GETTRANSACTIONSRESPONSE
DESCRIPTOR.message_types_by_name['GetTransactionTreesResponse'] = _GETTRANSACTIONTREESRESPONSE
DESCRIPTOR.message_types_by_name['GetTransactionByEventIdRequest'] = _GETTRANSACTIONBYEVENTIDREQUEST
DESCRIPTOR.message_types_by_name['GetTransactionByIdRequest'] = _GETTRANSACTIONBYIDREQUEST
DESCRIPTOR.message_types_by_name['GetTransactionResponse'] = _GETTRANSACTIONRESPONSE
DESCRIPTOR.message_types_by_name['GetFlatTransactionResponse'] = _GETFLATTRANSACTIONRESPONSE
DESCRIPTOR.message_types_by_name['GetLedgerEndRequest'] = _GETLEDGERENDREQUEST
DESCRIPTOR.message_types_by_name['GetLedgerEndResponse'] = _GETLEDGERENDRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetTransactionsRequest = _reflection.GeneratedProtocolMessageType('GetTransactionsRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONSREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetTransactionsRequest)
  ))
_sym_db.RegisterMessage(GetTransactionsRequest)

GetTransactionsResponse = _reflection.GeneratedProtocolMessageType('GetTransactionsResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONSRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetTransactionsResponse)
  ))
_sym_db.RegisterMessage(GetTransactionsResponse)

GetTransactionTreesResponse = _reflection.GeneratedProtocolMessageType('GetTransactionTreesResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONTREESRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetTransactionTreesResponse)
  ))
_sym_db.RegisterMessage(GetTransactionTreesResponse)

GetTransactionByEventIdRequest = _reflection.GeneratedProtocolMessageType('GetTransactionByEventIdRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONBYEVENTIDREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetTransactionByEventIdRequest)
  ))
_sym_db.RegisterMessage(GetTransactionByEventIdRequest)

GetTransactionByIdRequest = _reflection.GeneratedProtocolMessageType('GetTransactionByIdRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONBYIDREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetTransactionByIdRequest)
  ))
_sym_db.RegisterMessage(GetTransactionByIdRequest)

GetTransactionResponse = _reflection.GeneratedProtocolMessageType('GetTransactionResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETTRANSACTIONRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetTransactionResponse)
  ))
_sym_db.RegisterMessage(GetTransactionResponse)

GetFlatTransactionResponse = _reflection.GeneratedProtocolMessageType('GetFlatTransactionResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETFLATTRANSACTIONRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetFlatTransactionResponse)
  ))
_sym_db.RegisterMessage(GetFlatTransactionResponse)

GetLedgerEndRequest = _reflection.GeneratedProtocolMessageType('GetLedgerEndRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETLEDGERENDREQUEST,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetLedgerEndRequest)
  ))
_sym_db.RegisterMessage(GetLedgerEndRequest)

GetLedgerEndResponse = _reflection.GeneratedProtocolMessageType('GetLedgerEndResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETLEDGERENDRESPONSE,
  __module__ = 'com.digitalasset.ledger.api.v1.transaction_service_pb2'
  # @@protoc_insertion_point(class_scope:com.digitalasset.ledger.api.v1.GetLedgerEndResponse)
  ))
_sym_db.RegisterMessage(GetLedgerEndResponse)


DESCRIPTOR._options = None

_TRANSACTIONSERVICE = _descriptor.ServiceDescriptor(
  name='TransactionService',
  full_name='com.digitalasset.ledger.api.v1.TransactionService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=1551,
  serialized_end=2560,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetTransactions',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetTransactions',
    index=0,
    containing_service=None,
    input_type=_GETTRANSACTIONSREQUEST,
    output_type=_GETTRANSACTIONSRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetTransactionTrees',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetTransactionTrees',
    index=1,
    containing_service=None,
    input_type=_GETTRANSACTIONSREQUEST,
    output_type=_GETTRANSACTIONTREESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetTransactionByEventId',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetTransactionByEventId',
    index=2,
    containing_service=None,
    input_type=_GETTRANSACTIONBYEVENTIDREQUEST,
    output_type=_GETTRANSACTIONRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetTransactionById',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetTransactionById',
    index=3,
    containing_service=None,
    input_type=_GETTRANSACTIONBYIDREQUEST,
    output_type=_GETTRANSACTIONRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetFlatTransactionByEventId',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetFlatTransactionByEventId',
    index=4,
    containing_service=None,
    input_type=_GETTRANSACTIONBYEVENTIDREQUEST,
    output_type=_GETFLATTRANSACTIONRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetFlatTransactionById',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetFlatTransactionById',
    index=5,
    containing_service=None,
    input_type=_GETTRANSACTIONBYIDREQUEST,
    output_type=_GETFLATTRANSACTIONRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetLedgerEnd',
    full_name='com.digitalasset.ledger.api.v1.TransactionService.GetLedgerEnd',
    index=6,
    containing_service=None,
    input_type=_GETLEDGERENDREQUEST,
    output_type=_GETLEDGERENDRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_TRANSACTIONSERVICE)

DESCRIPTOR.services_by_name['TransactionService'] = _TRANSACTIONSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the definitions and implementations for expression types supported by the
query API.
"""

from typing import List, Optional, Set, Collection, Any

from .expr_base import Expr, exprify_args, exprify_value, TRIGGER_ANY_TIME, TRIGGER_ON_INIT
from ..model.types import Type, UnresolvedTypeReference


class UpdateStatementExpr(Expr):
    """ A statement in an update block. """

    def __getattr__(self, choice_name):
        """
        Generate a callable for the specified choice name. When the callable is invoked with
        arguments, an `ExerciseStatementExpr` is created that captures those args.
        """
        return _ExerciseStatementExprFactory(self, choice_name)

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, UpdateStatementExpr):
            return obj
        raise TypeError("Couldn't convert to a UpdateStatementExpr: {!r}".format(obj))


class TemplateExpr(Expr):
    """
    An :class:`Expr` wrapper around a :class:`TemplateRef`.

    In the context of an update block, this is coerced to a `create` statement.
    In the context of a `from` expression, this is coerced to a `TemplateSelectExpr` with a default
    label.
    """
    template_ref: Type

    def __init__(self, template_ref=None):
        self.template_ref = template_ref if isinstance(template_ref, Type) \
            else UnresolvedTypeReference(template_ref)

    @classmethod
    def coerce(cls, obj) -> 'TemplateExpr':
        """
        Attempt to coerce the specified type to a `TemplateExpr`.
        """
        if isinstance(obj, TemplateExpr):
            return obj
        if isinstance(obj, (Type, str)):
            return TemplateExpr(template_ref=obj)
        raise TypeError(f"Couldn't convert to a TemplateExpr: {obj!r}")

    def __getitem__(self, field_name: str) -> 'TemplateFieldExpr':
        """
        Return an expression that represents a field access of this template.
        """
        return TemplateFieldExpr(
            self, self._validate_field_name(field_name))

    def __getattr__(self, choice_name) -> '_TemplateChoiceFactory':
        return _TemplateChoiceFactory(self.template_ref, choice_name)

    def __call__(self, **kwargs):
        return CreateStatementExpr(template=self.template_ref, arguments=exprify_args(kwargs))

    def __repr__(self):
        return '<TemplateExpr({!r})>'.format(self.template_ref)

    def _validate_field_name(self, field_name):
        """
        Return the desired field name, or throw an error if the field name is unknown.

        The default implementation merely returns the exact field name passed in.
        """
        return field_name


class _TemplateChoiceFactory:
    __slots__ = ('template_ref', 'choice_name',)

    def __init__(self, template_ref, choice_name):
        self.template_ref = template_ref
        self.choice_name = choice_name

    def __call__(self, **kwargs) -> 'ExerciseStatementExpr':
        return ExerciseStatementExpr(
            expression=GetContractIdExpr(expression=TemplateSelectExpr(template=self.template_ref, alias=None)),
            choice_name=self.choice_name,
            arguments=exprify_args(kwargs))

    def __repr__(self):
        return f'<_TemplateChoiceFactory({self.choice_name!r})>'


class ConstantExpr(Expr):
    """ An expression that evaluates to a constant. """
    value: object

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, Expr):
            # TODO:
            return obj
        else:
            return ConstantExpr(value=obj)


class TemplateSelectExpr(Expr):
    """ A selection on a template. """
    template: TemplateExpr
    alias: str

    def __init__(self, template=None, alias=None):
        self.template = TemplateExpr.coerce(template)
        self.alias = alias

    @classmethod
    def coerce(cls, obj):
        """
        Attempt to coerce the specified type to a `TemplateSelectExpr`.
        """
        if isinstance(obj, TemplateSelectExpr):
            return obj
        if isinstance(obj, TemplateExpr):
            return TemplateSelectExpr(TemplateExpr.coerce(obj.template_ref), alias=None)
        elif isinstance(obj, str):
            return TemplateSelectExpr(TemplateExpr.coerce(obj), alias=None)
        raise TypeError("Couldn't convert to a {!r}: {!r}".format(cls.__name__, obj))

    def __repr__(self):
        return "<TemplateSelectExpr(template={!r}, alias={!r})>".format(self.template, self.alias)


class AppExpr(Expr):
    """
    An application of a function on one or arguments.
    """
    func: Any
    args: Collection[Expr]

    def __call__(self, *args, **kwargs):
        raise Exception(f'now where did THIS come from? args:{args} kwargs:{kwargs}')


class TemplateFieldExpr(Expr):
    """ An access of a field on a template. """
    expression: Expr
    field_name: str

    def __init__(self, expression=None, field_name=None):
        # coercion rules here are a little cumbersome:
        #  * str, TemplateRef, TemplateExpr and TemplateSelectExpr all expand to
        #    a cdata() reference on the appropriate contract fetch
        #  * all other expressions are assumed to be usable in this context
        # TODO: Formalize this here
        if isinstance(expression, (str, Type, TemplateExpr, TemplateSelectExpr)):
            expression = GetContractDataExpr(expression=TemplateSelectExpr.coerce(expression))
        elif not isinstance(expression, Expr):
            raise TypeError('Could not convert to an appropriate expression of a field access: {!r}'.format(expression))

        self.expression = expression
        self.field_name = field_name

    def __getitem__(self, sub_field_name):
        return TemplateFieldExpr(self, sub_field_name)

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, TemplateFieldExpr):
            return obj
        raise TypeError("Couldn't convert to a TemplateFieldExpr: {!r}".format(obj))


class TemplateSelectsExpr(Expr):
    """
    A set of template selections that define a cartesian join across all template options,
    paired with a predicate that establishes membership.
    """

    select: Set[TemplateSelectExpr]
    predicate: Optional[Expr]

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, TemplateSelectsExpr):
            return TemplateSelectsExpr(select=[TemplateSelectExpr.coerce(s) for s in obj.select], predicate=obj.predicate)
        elif isinstance(obj, TemplateSelectExpr):
            return TemplateSelectsExpr(select=[obj], predicate=None)
        elif isinstance(obj, TemplateExpr):
            return TemplateSelectsExpr(select=[TemplateSelectExpr(obj.template_ref, alias=None)])
        raise TypeError("Couldn't convert to a TemplateSelectsExpr: {!r}".format(obj))


class CreateStatementExpr(UpdateStatementExpr):
    """ A create statement. """
    template: Type
    arguments: dict


class ExerciseStatementExpr(UpdateStatementExpr):
    """ An exercise choice statement. """
    expression: Expr
    choice_name: str
    arguments: dict


class GetContractIdExpr(Expr):
    """
    Retrieve the contract ID for the specified expression. Automatically inserted at naked
    table references.
    """
    expression: TemplateSelectExpr

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, (Type, TemplateExpr, TemplateSelectExpr)):
            return GetContractIdExpr(expression=TemplateSelectExpr.coerce(obj))
        return super().coerce(obj)


class GetContractDataExpr(Expr):
    """
    Retrieve the contract data for the specified expression. Automatically inserted at field
    access sites.
    """
    expression: TemplateSelectExpr

    @classmethod
    def coerce(cls, obj):
        if isinstance(obj, (Type, TemplateExpr, TemplateSelectExpr)):
            return GetContractDataExpr(expression=TemplateSelectExpr.coerce(obj))


class _ExerciseStatementExprFactory:
    """
    A partially-applied choice exercise, containing everything except the actual arguments to
    parameters.
    """
    __slots__ = ('expression', 'choice_name')

    def __init__(self, expression, choice_name):
        self.expression = exprify_value(expression)
        self.choice_name = choice_name

    def __call__(self, **kwargs):
        return ExerciseStatementExpr(
            expression=self.expression,
            choice_name=self.choice_name,
            arguments=exprify_args(kwargs))

    def __repr__(self):
        return '<_ExerciseStatementExprFactory(expression={!r}, choice_name={!r})>'.format(
            self.expression, self.choice_name)


class UpdateBlockExpr(Expr):
    """ A block of updates. """
    statements: List[UpdateStatementExpr]

    def __init__(self, statements=None):
        self.statements = [UpdateStatementExpr.coerce(stmt) for stmt in _listify(statements)]

    @classmethod
    def coerce(cls, obj):
        """
        Attempt to coerce the specified type to a `UpdateBlockExpr`.
        """
        if obj is None:
            raise ValueError("can't be none")
        if isinstance(obj, UpdateBlockExpr):
            return obj
        if isinstance(obj, UpdateStatementExpr):
            return UpdateBlockExpr([obj])
        elif isinstance(obj, list):
            return UpdateBlockExpr(obj)
        raise TypeError("Couldn't convert to a UpdateBlockExpr: {!r}".format(obj))


class ProgramExpr(Expr):
    """
    A program, which is the combination of template selection with a block of updates and a trigger
    condition.
    """
    select: TemplateSelectsExpr
    update: UpdateBlockExpr
    trigger: str
    name: str

    def __init__(self, select=None, update=None, trigger=None, name=None):
        self.update = UpdateBlockExpr.coerce(update)
        self.trigger = trigger or TRIGGER_ANY_TIME
        self.name = name if name is not None else "p{}".format(id(self))

        if self.trigger == TRIGGER_ON_INIT and select is None:
            self.select = None
        elif select is not None:
            self.select = TemplateSelectsExpr.coerce(select)
        else:
            raise ValueError('a selection is required on ANY_TIME triggers')

    @staticmethod
    def coerce(obj):
        if isinstance(obj, ProgramExpr):
            return obj
        elif isinstance(obj, ExerciseStatementExpr):
            # only certain kinds of exercise statements can be easily converted to a Program
            # exercise 'choice_name' on cid(a) -> from a do update [ exercise 'choice_name' on cid(a) ]
            if isinstance(obj.expression, GetContractIdExpr) and \
               isinstance(obj.expression.expression, TemplateSelectExpr):
                return ProgramExpr(select=TemplateSelectsExpr.coerce(obj.expression.expression), update=UpdateBlockExpr.coerce(obj))
        raise TypeError("Couldn't convert to a ProgramExpr: {!r}".format(obj))


class SuiteExpr(Expr):
    """ A suite of initialization routines and programs. """
    party: str
    programs: List[ProgramExpr]


class Application(Expr):
    """
    An application, including the parties that are running programs on the ledger
    and required initialization steps.
    """
    suites: List[SuiteExpr]

    def as_string(self, color=False):
        """
        Print the `Application` expression tree.
        """
        from .visitor_pretty import pretty_print
        return pretty_print(self)

    def __str__(self):
        """
        Print the `Application` expression tree.
        """
        return self.as_string()


def and_(*predicates):
    """
    Convenience method for creating an AND expression over multiple expressions.
    """
    from .builtins import BUILTIN_AND
    return AppExpr(BUILTIN_AND, predicates)


def _listify(obj):
    """
    Wrap the object in a list if it is not already a list. `None` is converted into an empty list.
    """
    if obj is None:
        return []
    if isinstance(obj, str):
        return [obj]
    elif isinstance(obj, (list, tuple)):
        return obj
    else:
        return [obj]

